/* $EPIC:$ */
/*
 * log.h: header file for log.c
 *
 * Copyright  2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef __log_h__
#define __log_h__

/* You need to #include "nutty.h" before including this file */

typedef int	Logref;

	/* This is the general purpose logging API */

	/*
	 * Whenever you want to start a log, you should call this function
	 * to create a new log object.  The default state for a new log is
	 * "no filename", "log everything", "no rewrite", "no mangle", and
	 * "off".  You need to modify these log attributes _before_ you turn
	 * the log on.  You must at least set the filename.  The return value
	 * of this function is a handle by which you may refer to your log
	 * in future calls to this API.  You must at some point call 
	 * 'release_log()' for every value returned from this function.
	 */
	Logref	get_new_log_handle	(void);

	/*
	 * This function changes the "info" attributes of an active log.
	 * The attributes are explained in nutty.h, and detail just what
	 * you want logged.
	 */
	int	change_log_info		(Logref, const output_info *);

	/*
	 * This function creates an anonymous ircII function that will be
	 * executed for each line of output.  The return value of that
	 * ircII function will be substituted for the original output.
	 */
	int	change_log_rewrite	(Logref, const char *);

	/*
	 * This function sets the "mangler" flags, which determine how,
	 * if at all, the original output should be modified before it
	 * is committed to the file.  This value must have been returned
	 * by a prior call to "parse_mangle()" if you want the results
	 * to be meaningful.
	 */
	int	change_log_mangler	(Logref, int);

	/*
	 * This function changes the filename of this log.  If the log
	 * is already on, the current log will be closed and the new
	 * filename will be opened as a new log.
	 */
	int	change_log_filename	(Logref, const char *);

	/*
	 * After you have set all of the neccesary attributes, you may call
	 * this function to actually begin logging.  Logging will continue
	 * until you call turn_log_off(), release_log(), or the client exits.
	 */
	int	turn_log_on		(Logref);

	/*
	 * When you want to suspend logging, you may call this function.
	 * No logging will occur, but the log remains active, after this
	 * function is called.  You may begin logging again by calling
	 * "turn_log_on"
	 */
	int	turn_log_off		(Logref);

	/*
	 * When you are completely done with a log, you must call this
	 * function.  The value passed to this function must have been
	 * previously returned by "get_new_log_handle()".  When you call
	 * this function, THE LOG IS REVOKED AND WILL BE RE-USED and any 
	 * attempt to operate on the log will yeild an error, or worse 
	 * yet, you might modify someonbody else's log, a truly rude act.
	 * As every active log requires handling for each line of output
	 * in the *ENTIRE* client, you should release logs when you are
	 * done, to be gracious to the user.
	 */
	int	release_log		(Logref);


	/* Other functions */
	/* This is called by the exit routines */
	void	cancel_all_logs		(void);

	/* These are callbacks for /SET LOGFILE and /SET LOG */
	void	set_logfile_callback	(const char *);
	void	set_log_callback	(int);
	void	set_mangle_logfiles	(const char *);

	BUILT_IN_COMMAND(logcmd);

#endif
