/* $EPIC:$ */
/*
 * alias-assign.c -- Handles the whole kit and caboodle for all variables
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1997, 2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */
/* Split from alias.c in Jan 2000 */

#include "irc.h"
#include "alias.h"
#include "alias-private.h"
#include "alist.h"
#include "commands.h"
#include "ircaux.h"
#include "output.h"
#include "vars.h"
#include "symtab.h"

/*
 * after_expando: This replaces some much more complicated logic strewn
 * here and there that attempted to figure out just how long an expando 
 * name was supposed to be.  Well, now this changes that.  This will slurp
 * up everything in 'start' that could possibly be put after a $ that could
 * result in a syntactically valid expando.  All you need to do is tell it
 * if the expando is an rvalue or an lvalue (it *does* make a difference)
 */
static 	char *lval[] = { "rvalue", "lvalue" };
char *	after_expando (const char *start, int lvalue, int *call)
{
	char	*rest;
	char	*str;

	if (!*start)
		return (char *)start;		/* Oh hush */

	/*
	 * One or two leading colons are allowed
	 */
	str = (char *)start;
	if (*str == ':')
		if (*++str == ':')
			++str;

	/*
	 * This handles 99.99% of the cases
	 */
	while (*str && (isalpha(*str) || isdigit(*str) || 
				*str == '_' || *str == '.'))
		str++;

	/*
	 * This handles any places where a var[var] could sneak in on
	 * us.  Supposedly this should never happen, but who can tell?
	 */
	while (*str == '[')
	{
		if (!(rest = MatchingBracket(str + 1, '[', ']')))
		{
			if (!(rest = strchr(str, ']')))
			{
				yell("Unmatched bracket in %s (%s)", 
						lval[lvalue], start);
				return empty_string;
			}
		}
		str = rest + 1;
	}

	/*
	 * Rvalues may include a function call, slurp up the argument list.
	 */
	if (!lvalue && *str == '(')
	{
		if (!(rest = MatchingBracket(str + 1, '(', ')')))
		{
			if (!(rest = strchr(str, ')')))
			{
				yell("Unmatched paren in %s (%s)", 
						lval[lvalue], start);
				return empty_string;
			}
		}
		*call = 1;
		str = rest + 1;
	}

	/*
	 * If the entire thing looks to be invalid, perhaps its a 
	 * special built-in expando.  Check to see if it is, and if it
	 * is, then slurp up the first character as valid.
	 * Also note that $: by itself must be valid, which requires
	 * some shenanigans to handle correctly.  Ick.
	 */
	if (str == start || (str == start + 1 && *start == ':'))
	{
		int is_builtin = 0;

		built_in_alias(*start, &is_builtin);
		if (is_builtin && (str == start))
			str++;
	}


	/*
	 * All done!
	 */
	return str;		/* Oh hush. */
}

/* Used for statistics gathering */
extern unsigned long 	alias_total_allocated; 
extern unsigned long 	alias_total_bytes_allocated;

/*
 * This is where we keep track of where the last pending function call.
 * This is used when you assign to the FUNCTION_RETURN value.  Since it
 * is neccesary to be able to access FUNCTION_RETURN in contexts where other
 * local variables would not be visible, we do this as a quasi-neccesary
 * hack.  When you reference FUNCTION_RETURN, it goes right to this stack.
 */
	int	last_function_call_level = -1;

/*
 * The following actions are supported:  add, delete, find, list
 * On the following types of data:	 var_alias, local_alias
 * Except you cannot list or delete local_aliases.
 *
 * To fetch a variable, use ``get_variable''
 * To recurse an array structure, use ``get_subarray_elements''
 */

/*
 * These are general purpose interface functions.
 * However, the static ones should **always** be static!  If you are tempted
 * to use them outside of alias.c, please rethink what youre trying to do.
 * Using these functions violates the encapsulation of the interface.
 * Specifically, if you create a global variable and then want to delete it,
 * try using a local variable so it is reaped automatically.
 */
static	void	delete_var_alias   (char *name, int noisy);
static	void	list_var_alias     (char *name);
static	void	list_local_alias   (char *name);

/************************** HIGH LEVEL INTERFACE ***********************/
/*
 * User front end to the ASSIGN command
 * Syntax to add variable:    ASSIGN name text
 * Syntax to delete variable: ASSIGN -name
 */
BUILT_IN_COMMAND(assigncmd)
{
	char *real_name;
	char *name;

	/*
	 * If there are no arguments, list all the global variables
	 */
	if (!(name = next_arg(args, &args)))
	{
		list_var_alias(NULL);
		return;
	}

	/*
	 * Canonicalize the variable name
	 */
	real_name = remove_brackets(name, NULL, NULL);

	/*
	 * Find the stuff to assign to the variable
	 */
	while (my_isspace(*args))
		args++;

	/*
	 * If there is no body, then the user probably wants to delete
	 * the variable
	 */
	if (!args || !*args)
	{
		/*
	 	 * If the variable name starts with a hyphen, then we remove
		 * the variable
		 */
		if (real_name[0] == '-')
		{
			if (real_name[1])
				delete_var_alias(real_name + 1, 1);
			else
				say("You must specify an alias to be removed.");
		}

		/*
		 * Otherwise, the user wants us to list the variable
		 */
		else
			list_var_alias(real_name);
	}

	/*
	 * Register the variable
	 */
	else
		add_var_alias(real_name, args, 1);

	new_free(&real_name);
	return;
}

BUILT_IN_COMMAND(localcmd)
{
	char *name;

	if (!(name = next_arg(args, &args)))
	{
		list_local_alias(NULL);
		return;
	}

	while (args && *args && my_isspace(*args))
		args++;

	if (!args)
		args = empty_string;

	while (name && *name)
	{
		char 	*next_name;
		char	*real_name;

		if ((next_name = strchr(name, ',')))
			*next_name++ = 0;

		real_name = remove_brackets(name, NULL, NULL);
		add_local_alias(real_name, args, 1);
		new_free(&real_name);
		name = next_name;
	}
}

/**************************** INTERMEDIATE INTERFACE *********************/
	RuntimeStack *	call_stack = NULL;
	int 		max_wind = -1;
	int 		wind_index = -1;

static	Alias *	check_var_alias_unstub (Alias *item);


/***************************************************************************/
/*
 * add_var_alias: Add a global variable
 *
 * name -- name of the alias to create (must be canonicalized)
 * stuff -- what to have ``name'' expand to.
 *
 * If ``name'' is FUNCTION_RETURN, then it will create the local
 * return value (die die die)
 *
 * If ``name'' refers to an already created local variable, that
 * local variable is used (invisibly)
 */
void	add_var_alias	(const char *name1, const char *stuff, int noisy)
{
	char 	*ptr;
	int 	af;
	int	local = 0;
	char	*save;
	char 	*name;

	save = name = remove_brackets(name1, NULL, &af);
	if (*name == ':')
	{
		name++, local = 1;
		if (*name == ':')
			name++, local = -1;
	}

	/*
	 * Weed out invalid variable names
	 */
	ptr = after_expando(name, 1, NULL);
	if (*ptr)
		error("ASSIGN names may not contain '%c' "
			"(You asked for [%s])", *ptr, name);

	/*
	 * Weed out FUNCTION_RETURN (die die die)
	 */
	else if (!strcmp(name, "FUNCTION_RETURN"))
		add_local_alias(name, stuff, noisy);

	/*
	 * Pass the buck on local variables
	 */
	else if ((local == 1) || (local == 0 && find_local_alias(name, NULL)))
		add_local_alias(name, stuff, noisy);

	else if (stuff && *stuff)
	{
		Alias 	*tmp = find_var_alias(name, 1);

		/*
		 * Then we fill in the interesting details
		 */
		if (!tmp->filename || strcmp(tmp->filename, current_package()))
			malloc_strcpy(&tmp->filename, current_package());
		malloc_strcpy(&tmp->stuff, stuff);
		new_free(&tmp->stub);
		tmp->global = loading_global;

		alias_total_allocated++;
		alias_total_bytes_allocated += strlen(tmp->name) + 
			strlen(tmp->stuff) + strlen(tmp->filename);

		/*
		 * And tell the user.
		 */
		if (noisy)
			say("Assign %s added [%s]", name, stuff);
	}
	else
		delete_var_alias(name, noisy);

	new_free(&save);
	return;
}

void	add_var_stub_alias  (Char *name, Char *stuff, int noisy)
{
	char *	ptr;
	int	af;

	name = remove_brackets(name, NULL, &af);

	ptr = after_expando(name, 1, NULL);
	if (*ptr)
		error("Assign names may not contain '%c' "
			"(You asked for [%s])", *ptr, name);

	else if (!strcmp(name, "FUNCTION_RETURN"))
		error("You may not stub the FUNCTION_RETURN variable.");

	else 
	{
		Alias *	tmp = find_var_alias(name, 1);

		if (!tmp->filename || strcmp(tmp->filename, current_package()))
			malloc_strcpy(&tmp->filename, current_package());
		malloc_strcpy(&tmp->stub, stuff);
		new_free(&tmp->stuff);
		tmp->global = loading_global;

		alias_total_allocated++;
		alias_total_bytes_allocated += strlen(tmp->stub);
		if (noisy)
			say("Assign %s stubbed to file %s", name, stuff);
	}

	new_free(&name);
	return;
}

/*
 * This is the master function to retreive an assign alias by name.
 * "name" is expected to be in canonical form (all uppercase, dot notation)
 * and "create" is 1 if you want the assign to be created if it does not 
 * already exist.
 */
Alias *	find_var_alias (const char *name, int create)
{
	vt_item	*	item = NULL;
	Alias *		a = NULL;

	if (!strncmp(name, "::", 2))
		name += 2;		/* Accept ::global */

	if (!(item = get_var_item(name)) || !item->assign)
	{
		if (create)
		{
			a = make_new_Alias(name);
			add_assign_var_entry(name, a);
			return a;
		}
		else
			return NULL;
	}

	a = (Alias *)(item->assign);
	if (a->stub)
		a = check_var_alias_unstub(a);

	if (a && !a->stuff)
		panic("a->stuff is NULL and it shouldn't be.");

        /* Take *THAT* */
        if (!a && create)
                return find_var_alias(name, create);

	return a;
}

static Alias *	check_var_alias_unstub (Alias *item)
{
	char *	file;
	char *	name;

	file = LOCAL_COPY(item->stub);
	name = LOCAL_COPY(item->name);
	delete_var_alias(name, 0);

	item = NULL;
	if (!unstub_in_progress)
	{
		unstub_in_progress = 1;
		load("LOAD", file, empty_string);
		unstub_in_progress = 0;
	}

	/*
	 * At this opint, we have to see if 'item' was
	 * redefinied by the /load.  We call find_var_alias
	 * recursively to pick up the new value.
	 */
	return find_var_alias(name, 0);
}

/*
 * Delete a global assign alias by name.
 */
static void	delete_var_alias (char *name, int noisy)
{
	Alias *	item;
	char *	name_copy;

	name_copy = LOCAL_COPY(name);
	name = name_copy;

	upper(name);
	if ((item = (Alias *)remove_assign_var_entry(name)))
	{
		new_free(&item->name);
		new_free(&item->stuff);
		new_free(&item->stub);
		new_free(&item->filename);
		new_free((char **)&item);
		if (noisy)
			say("Assign %s removed", name);
	}
	else if (noisy)
		say("No such assign: %s", name);
}

/*
 * This function is strictly O(N).  Its possible to address this.
 */
static void	list_var_alias (char *name)
{
	int		count;
	void *		ptr;
	Alias *		a;
	int		len;
	size_t		first_dot,
			curr_struct_len = 0;
	const char *	curr_struct = NULL;
	const char *	s;
	const char *	script;

	if (name)
	{
		upper(name);
		len = strlen(name);
	}
	else
		len = 0;

	say("Assigns:");
	for (count = 0; (ptr = foreach_assign_item(&count)); count++)
	{
		a = (Alias *)ptr;

		if (name)
		{
			int	c = strncmp(a->name, name, len);
			if (c < 0)
				continue;
			if (c > 0)
				break;
		}

		script = *a->filename ? a->filename : "*";

		if ((s = strchr(a->name + len, '.')))
		{
			first_dot = s - a->name;

			if (!curr_struct ||
			    (first_dot != curr_struct_len) || 
			    strncmp(a->name, curr_struct, first_dot))
			{
				say("[%s]\t%*.*s\t<Structure>", 
					script, (int)first_dot, 
					(int)first_dot, a->name);
				curr_struct = a->name;
				curr_struct_len = first_dot;
			}
		}
		else
		{
			if (a->stub)
				say("[%s]\t%s STUBBED TO %s", 
					script, a->name, a->stub);
			else
				say("[%s]\t%s\t%s", 
					script, a->name, a->stuff);
		}
	}
}

void	unload_var_alias (const char *filename)
{
	int 	cnt;
	void *	ptr;
	Alias *	a;

	for (cnt = 0; (ptr = foreach_assign_item(&cnt)); cnt++)
	{
		a = (Alias *)ptr;
		if (!strcmp(a->filename, filename))
			delete_var_alias(a->name, 0);
	}
}

/*
 * This function is strictly O(N).  This should probably be addressed.
 */
char **	glob_assign_alias (char *name, int *howmany)
{
	int 		cnt;
	void *		ptr;
	Alias *		a;
	int     	cmp;
	int     	len;
	char **		matches = NULL;
	int     	matches_size = 5;

	len = strlen(name);
	*howmany = 0;
	matches = RESIZE(matches, char *, matches_size);

	for (cnt = 0; (ptr = foreach_assign_item(&cnt)); cnt++)
	{
		a = (Alias *)ptr;

		if (!(cmp = strncmp(name, a->name, len)))
		{
			if (strchr(a->name + len, '.'))
				continue;

			matches[*howmany] = m_strdup(a->name);
			(*howmany)++;
			if (*howmany == matches_size)
			{
				matches_size += 5;
				RESIZE(matches, char *, matches_size);
			}
		}
		else if (cmp < 0)
			break;
	}

	if (*howmany)
		matches[*howmany] = NULL;
	else
		new_free((char **)&matches);

	return matches;
}

/*
 * This function is strictly O(N).  This should probably be addressed.
 */
char **	pmatch_assign_alias (char *name, int *howmany)
{
	int    		cnt;
	void *		ptr;
	Alias *		a;
	int     	len;
	char **		matches = NULL;
	int     	matches_size = 5;

	len = strlen(name);
	*howmany = 0;
	matches = RESIZE(matches, char *, matches_size);

	for (cnt = 0; (ptr = foreach_assign_item(&cnt)); cnt++)
	{
		a = (Alias *)ptr;

		if (wild_match(name, a->name))
		{
			matches[*howmany] = m_strdup(a->name);
			*howmany += 1;
			if (*howmany == matches_size)
			{
				matches_size += 5;
				RESIZE(matches, char *, matches_size);
			}
		}
	}

	if (*howmany)
		matches[*howmany] = NULL;
	else
		new_free((char **)&matches);

	return matches;
}

/*
 * save_aliases: This will write all of the aliases to the FILE pointer fp in
 * such a way that they can be read back in using LOAD or the -l switch 
 */
void 	save_assigns	(FILE *fp, int do_all)
{
	int	cnt = 0;
	void *	ptr;
	Alias *	a;

	for (cnt = 0; (ptr = foreach_assign_item(&cnt)); cnt++)
	{
		a = (Alias *)ptr;
		if (!a->global || do_all)
		{
			if (a->stub)
				fprintf(fp, "STUB ASSIGN %s %s", 
					a->name, a->stub);
			else
				fprintf(fp, "ASSIGN %s %s\n", 
					a->name, a->stuff);
		}
	}
}



/************************ LOCAL VARIABLES ********************************/
/*
 * add_local_alias: Add a local variable to the current run stack.
 * Note that unlike global variables, local variables are put in a stack
 * that is completely seperate from everything else.  This is a hack and
 * probably should be changed, but for now it works.
 */
void	add_local_alias	(Char *name, Char *stuff, int noisy)
{
	char 	*ptr;
	Alias 	*tmp = NULL;
	AliasSet *list = NULL;
	int 	af;

	name = remove_brackets(name, NULL, &af);

	/*
	 * Weed out invalid variable names
	 */
	ptr = after_expando(name, 1, NULL);
	if (*ptr)
	{
		error("LOCAL names may not contain '%c' (You asked for [%s])", 
						*ptr, name);
		new_free(&name);
		return;
	}
	
	/*
	 * Now we see if this local variable exists anywhere
	 * within our view.  If it is, we dont care where.
	 * If it doesnt, then we add it to the current frame,
	 * where it will be reaped later.
	 */
	if (!(tmp = find_local_alias (name, &list)))
	{
		tmp = make_new_Alias(name);
		add_to_array ((array *)list, (array_item *)tmp);
	}

	/* Fill in the interesting stuff */
	if (!tmp->filename || strcmp(tmp->filename, current_package()))
		malloc_strcpy(&(tmp->filename), current_package());
	malloc_strcpy(&(tmp->stuff), stuff);
	alias_total_allocated++;
	if (tmp->stuff)		/* Oh blah. */
	{
		alias_total_bytes_allocated += strlen(tmp->stuff);
		if (x_debug & DEBUG_LOCAL_VARS)
		    yell("Assign %s (local) added [%s]", name, stuff);
		else if (noisy)
		    say("Assign %s (local) added [%s]", name, stuff);
	}

	new_free(&name);
	return;
}

/*
 * An example will best describe the semantics:
 *
 * A local variable will be returned if and only if there is already a
 * variable that is exactly ``name'', or if there is a variable that
 * is an exact leading subset of ``name'' and that variable ends in a
 * period (a dot).
 */
Alias *	find_local_alias (const char *name, AliasSet **list)
{
	Alias 	*alias = NULL;
	int 	c = wind_index;
	char 	*ptr;
	int 	implicit = -1;
	int	function_return = 0;

	/* No name is an error */
	if (!name)
		return NULL;

	ptr = after_expando(name, 1, NULL);
	if (*ptr)
		return NULL;

	if (!my_stricmp(name, "FUNCTION_RETURN"))
		function_return = 1;

	/*
	 * Search our current local variable stack, and wind our way
	 * backwards until we find a NAMED stack -- that is the enclosing
	 * alias or ON call.  If we find a variable in one of those enclosing
	 * stacks, then we use it.  If we dont, we progress.
	 *
	 * This needs to be optimized for the degenerate case, when there
	 * is no local variable available...  It will be true 99.999% of
	 * the time.
	 */
	for (c = wind_index; c >= 0; c = call_stack[c].parent)
	{
		/* XXXXX */
		if (function_return && last_function_call_level != -1)
			c = last_function_call_level;

		if (x_debug & DEBUG_LOCAL_VARS)
			yell("Looking for [%s] in level [%d]", name, c);

		if (call_stack[c].alias.list)
		{
		    int x;

		    /* XXXX - This is bletcherous */
		    for (x = 0; x < call_stack[c].alias.max; x++)
		    {
			size_t	len;

			len = strlen(call_stack[c].alias.list[x]->name);
			if (streq(call_stack[c].alias.list[x]->name, name) 
					== len)
			{
			    if (call_stack[c].alias.list[x]->name[len-1] == '.')
				implicit = c;
			    else if (strlen(name) == len)
			    {
				alias = call_stack[c].alias.list[x];
				break;
			    }
			}
			else
			{
			    if (my_stricmp(call_stack[c].alias.list[x]->name, 
						name) > 0)
				continue;
			}
		     }

		    if (!alias && implicit >= 0)
		    {
			alias = make_new_Alias(name);
			add_to_array((array *)&call_stack[implicit].alias, 
					(array_item *)alias);
		    }
		}

		if (alias)
		{
			if (x_debug & DEBUG_LOCAL_VARS) 
				yell("I found [%s] in level [%d] (%s)", 
					name, c, alias->stuff);
			break;
		}

		if (function_return)			/* XXX May 8, 2000 */
			break;				/* XXX May 8, 2000 */

		if (*call_stack[c].name || call_stack[c].parent == -1)
		{
			if (x_debug & DEBUG_LOCAL_VARS) 
			    yell("I didnt find [%s], stopped at level [%d]", 
				name, c);
			break;
		}
	}

	if (alias)
	{
		if (list)
			*list = &call_stack[c].alias;
		return alias;
	}
	else if (list)
		*list = &call_stack[wind_index].alias;

	return NULL;
}


static void	list_local_alias (char *name)
{
	int len = 0, cnt;
	int DotLoc, LastDotLoc = 0;
	char *LastStructName = NULL;
	char *s;

	say("Visible Local Assigns:");
	if (name)
		len = strlen(upper(name));

	for (cnt = wind_index; cnt >= 0; cnt = call_stack[cnt].parent)
	{
		int x;
		if (!call_stack[cnt].alias.list)
			continue;
		for (x = 0; x < call_stack[cnt].alias.max; x++)
		{
			if (!name || !strncmp(call_stack[cnt].alias.list[x]->name, name, len))
			{
				if ((s = strchr(call_stack[cnt].alias.list[x]->name + len, '.')))
				{
					DotLoc = s - call_stack[cnt].alias.list[x]->name;
					if (!LastStructName || (DotLoc != LastDotLoc) || strncmp(call_stack[cnt].alias.list[x]->name, LastStructName, DotLoc))
					{
						put_it("\t%*.*s\t<Structure>", DotLoc, DotLoc, call_stack[cnt].alias.list[x]->name);
						LastStructName = call_stack[cnt].alias.list[x]->name;
						LastDotLoc = DotLoc;
					}
				}
				else
					put_it("\t%s\t%s", call_stack[cnt].alias.list[x]->name, call_stack[cnt].alias.list[x]->stuff);
			}
		}
	}
}

void 	make_local_stack 	(const char *name)
{
	wind_index++;

	if (wind_index >= max_wind)
	{
		int tmp_wind = wind_index;

		if (max_wind == -1)
			max_wind = 8;
		else
			max_wind <<= 1;

		RESIZE(call_stack, RuntimeStack, max_wind);
		for (; wind_index < max_wind; wind_index++)
		{
			call_stack[wind_index].alias.max = 0;
			call_stack[wind_index].alias.max_alloc = 0;
			call_stack[wind_index].alias.list = NULL;
			call_stack[wind_index].alias.func = strncmp;
			call_stack[wind_index].current = NULL;
			call_stack[wind_index].name = NULL;
			call_stack[wind_index].parent = -1;
		}
		wind_index = tmp_wind;
	}

	/* Just in case... */
	destroy_local_stack();
	wind_index++;		/* XXXX - chicanery */

	if (name)
	{
		call_stack[wind_index].name = name;
		call_stack[wind_index].parent = -1;
	}
	else
	{
		call_stack[wind_index].name = empty_string;
		call_stack[wind_index].parent = wind_index - 1;
	}
	call_stack[wind_index].locked = 0;
}

int	find_locked_stack_frame	(void)
{
	int i;
	for (i = 0; i < wind_index; i++)
		if (call_stack[i].locked)
			return i;

	return -1;
}

void	bless_local_stack	(void)
{
	call_stack[wind_index].name = empty_string;
	call_stack[wind_index].parent = find_locked_stack_frame();
}

void 	destroy_local_stack 	(void)
{
	/*
	 * We clean up as best we can here...
	 */
	if (call_stack[wind_index].alias.list)
		destroy_aliases(VAR_ALIAS_LOCAL);
	if (call_stack[wind_index].current)
		call_stack[wind_index].current = 0;
	if (call_stack[wind_index].name)
		call_stack[wind_index].name = 0;

	wind_index--;
}

void 	set_current_command 	(char *line)
{
	call_stack[wind_index].current = line;
}

void 	unset_current_command 	(void)
{
	call_stack[wind_index].current = NULL;
}

void	lock_stack_frame 	(void)
{
	call_stack[wind_index].locked = 1;
}

void	unlock_stack_frame	(void)
{
	int lock = find_locked_stack_frame();
	if (lock >= 0)
		call_stack[lock].locked = 0;
}

void 	dump_call_stack 	(void)
{
	int my_wind_index = wind_index;
	if (wind_index >= 0)
	{
		say("Call stack");
		while (my_wind_index--)
			say("[%3d] %s", my_wind_index, call_stack[my_wind_index].current);
		say("End of call stack");
	}
}

void 	panic_dump_call_stack 	(void)
{
	int	my_wind_index = wind_index;

	fprintf(stderr, "Call stack\n");
	if (wind_index >= 0)
	{
		while (my_wind_index--)
			fprintf(stderr, "[%3d] %s\n", my_wind_index, 
					call_stack[my_wind_index].current);
	}
	else
		fprintf(stderr, "Stack is corrupted [wind_index is %d], sorry.\n",
			wind_index);
	fprintf(stderr, "End of call stack\n");
}


/*
 * You may NOT call this unless youre about to exit.
 * If you do (call this when youre not about to exit), and you do it more 
 * than a few times, max_wind will get absurdly large.  So dont do it.
 *
 * XXXX - this doesnt clean up everything -- but do i care?
 */
void 	destroy_call_stack 	(void)
{
	wind_index = 0;
	new_free((char **)&call_stack);
}



/************************* DIRECT VARIABLE EXPANSION ************************/
/*
 * get_variable: This simply looks up the given str.  It first checks to see
 * if its a user variable and returns it if so.  If not, it checks to see if
 * it's an IRC variable and returns it if so.  If not, it checks to see if
 * its and environment variable and returns it if so.  If not, it returns
 * null.  It mallocs the returned string 
 */
char 	*get_variable 	(const char *str)
{
	int af;
	return get_variable_with_args(str, NULL, &af);
}


char	*get_variable_with_args (const char *str, const char *args, int *args_flag)
{
	Alias	*alias = NULL;
	char	*ret = NULL;
	char	*name = NULL;
	char	*freep = NULL;
	int	copy = 0;
	int	local = 0;

	freep = name = remove_brackets(str, args, args_flag);

	/*
	 * Support $:var to mean local variable ONLY (no globals)
	 * Support $::var to mean global variable ONLY (no locals)
	 * Support $: with nothing after it as a built in expando.  Ick.
	 */
	if (*name == ':' && name[1])
	{
		name++, local = 1;
		if (*name == ':')
			name++, local = -1;
	}

	/*
	 * local == -1  means "local variables not allowed"
	 * local == 0   means "locals first, then globals"
	 * local == 1   means "global variables not allowed"
	 */
	if ((local != -1) && (alias = find_local_alias(name, NULL)))
		copy = 1, ret = alias->stuff;
	else if (local == 1)
		;

#if 0
	else if ((item = get_var_item(name)))
	{
		if (item->assign)
			ret = call_user_assign(item)
		else if (item->builtin)
			ret = call_builtin_assign(item->builtin);
		else if (item->set)
			ret = get_set_as_string(item->set);
		else if (item->environ)
			ret = m_strdup((const char *)item->environ);
	}
	else if (x_debug & DEBUG_UNKNOWN && ret == NULL)
		yell("Variable lookup to non-existant assign [%s]", name);
	new_free(&freep);
	return ret;
#else
	else if ((alias = find_var_alias(name, 0)) != NULL)
		copy = 1, ret = alias->stuff;
	else if ((strlen(str) == 1) && (ret = built_in_alias(*str, NULL)))
		copy = 0;
	else if ((ret = make_string_var(str)))
		copy = 0;
	else
		copy = 1, ret = getenv(str);

	if (x_debug & DEBUG_UNKNOWN && ret == NULL)
		yell("Variable lookup to non-existant assign [%s]", name);

	new_free(&freep);
	return (copy ? m_strdup(ret) : ret);
#endif
}



