/* $EPIC:$ */
/*
 * alias-command.c -- Handles the whole kit and caboodle for command aliases.
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1997, 1998 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */
/* Almost totaly rewritten by Jeremy Nelson (01/97) */
/* Split from alias.c in Jan 2000 */

#include "irc.h"
#include "alias.h"
#include "alias-private.h"
#include "alist.h"
#include "commands.h"
#include "ircaux.h"
#include "output.h"
#include "symtab.h"

#define LEFT_BRACE '{'
#define RIGHT_BRACE '}'
#define LEFT_BRACKET '['
#define RIGHT_BRACKET ']'
#define LEFT_PAREN '('
#define RIGHT_PAREN ')'
#define DOUBLE_QUOTE '"'

	/* Used for statistics gathering */
	u_long 		alias_total_allocated = 0;
	u_long 		alias_total_bytes_allocated = 0;

static	ArgList	*	parse_arglist (char *arglist);
static	void		destroy_arglist (ArgList *);

	int		unstub_in_progress;

/*
 * The following actions are supported:  add, delete, find, list
 * On the following types of data:	 var_alias, cmd_alias, local_alias
 * Except you cannot list or delete local_aliases.
 *
 * To fetch an ambiguous alias, use ``glob_cmd_alias''
 */

/*
 * These are general purpose interface functions.
 * However, the static ones should **always** be static!  If you are tempted
 * to use them outside of alias.c, please rethink what youre trying to do.
 * Using these functions violates the encapsulation of the interface.
 * Specifically, if you create a global variable and then want to delete it,
 * try using a local variable so it is reaped automatically.
 */
extern  void    add_cmd_stub_alias (char *name, char *stuff, int noisy);
static Alias *	check_cmd_alias_unstub (Alias *item, int *count);
static	void	list_cmd_alias     (char *name);

/************************** HIGH LEVEL INTERFACE ***********************/
/*
 * User front end to the ALIAS command
 * Syntax to debug alias:  ALIAS /S
 * Syntax to add alias:    ALIAS name [{] irc command(s) [}]
 * Syntax to delete alias: ALIAS -name
 */
BUILT_IN_COMMAND(aliascmd)
{
	char *name;
	char *real_name;
	char *ptr;

	/*
	 * If no name present, list all aliases
	 */
	if (!(name = next_arg(args, &args)))
	{
		list_cmd_alias(NULL);
		return;
	}

	/*
	 * Alias can take an /s arg, which shows some data we've collected
	 */
	if (!my_strnicmp(name, "/S", 2))
	{
	    say("Total aliases handled: %ld", 
			alias_total_allocated);
	    say("Total bytes allocated to aliases: %ld", 
			alias_total_bytes_allocated);
	    command_table_stats();
	    return;
	}

	/*
	 * Canonicalize the alias name
	 */
	real_name = remove_brackets(name, NULL, NULL);

	/*
	 * Find the argument body
	 */
	while (my_isspace(*args))
		args++;

	/*
	 * If there is no argument body, we probably want to delete alias
	 */
	if (!args || !*args)
	{
		/*
		 * If the alias name starts with a hyphen, then we are
		 * going to delete it.
		 */
		if (real_name[0] == '-')
		{
			if (real_name[1])
				delete_cmd_alias(real_name + 1, 1);
			else
				say("You must specify an alias to be removed.");
		}

		/*
		 * Otherwise, the user wants us to list that alias
		 */
		else
			list_cmd_alias(real_name);
	}

	/*
	 * If there is an argument body, then we have to register it
	 */
	else
	{
		ArgList *arglist = NULL;

		/*
		 * Aliases may contain a parameter list, which is parsed
		 * at registration time.
		 */
		if (*args == LEFT_PAREN)
		{
			ptr = MatchingBracket(++args, LEFT_PAREN, RIGHT_PAREN);
			if (!ptr)
				say("Unmatched lparen in %s %s", 
					command, real_name);
			else
			{
				*ptr++ = 0;
				while (*ptr && my_isspace(*ptr))
					ptr++;
				if (!*ptr)
					say("Missing alias body in %s %s", 
						command, real_name);

				while (*args && my_isspace(*args))
					args++;

				arglist = parse_arglist(args);
				args = ptr;
			}
		}

		/*
		 * Aliases' bodies can be surrounded by a set of braces,
		 * which are stripped off.
		 */
		if (*args == LEFT_BRACE)
		{
			ptr = MatchingBracket(++args, LEFT_BRACE, RIGHT_BRACE);

			if (!ptr)
				say("Unmatched brace in %s %s", 
						command, real_name);
			else 
			{
				*ptr++ = 0;
				while (*ptr && my_isspace(*ptr))
					ptr++;

				if (*ptr)
					say("Junk [%s] after closing brace "
						"in %s %s", ptr, command, 
						real_name);

				while (*args && my_isspace(*args))
					args++;

			}
		}

		/*
		 * Register the alias
		 */
		add_cmd_alias(real_name, arglist, args, 1);
	}

	new_free(&real_name);
	return;
}


/*
 * Argument lists look like this:
 *
 * LIST   := LPAREN TERM [COMMA TERM] RPAREN)
 * LPAREN := '('
 * RPAREN := ')'
 * COMMA  := ','
 * TERM   := LVAL QUAL | "..." | "void"
 * LVAL   := <An alias name>
 * QUAL   := NUM "words"
 *
 * In English:
 *   An argument list is a comma seperated list of variable descriptors (TERM)
 *   enclosed in a parenthesis set.  Each variable descriptor may contain any
 *   valid alias name followed by an action qualifier, or may be the "..."
 *   literal string, or may be the "void" literal string.  If a variable
 *   descriptor is an alias name, then that positional argument will be removed
 *   from the argument list and assigned to that variable.  If the qualifier
 *   specifies how many words are to be taken, then that many are taken.
 *   If the variable descriptor is the literal string "...", then all argument
 *   list parsing stops there and all remaining alias arguments are placed 
 *   into the $* special variable.  If the variable descriptor is the literal
 *   string "void", then the balance of the remaining alias arguments are 
 *   discarded and $* will expand to the false value.  If neither "..." nor
 *   "void" are provided in the argument list, then that last variable in the
 *   list will recieve all of the remaining arguments left at its position.
 * 
 * Examples:
 *
 *   # This example puts $0 in $channel, $1 in $mag, and $2- in $nicklist.
 *   /alias opalot (channel, mag, nicklist) {
 *	fe ($nicklist) xx yy zz {
 *	    mode $channel ${mag}ooo $xx $yy $zz
 *	}
 *   }
 *
 *   # This example puts $0 in $channel, $1 in $mag, and the new $* is old $2-
 *   /alias opalot (channel, mag, ...) {
 *	fe ($*) xx yy zz {
 *	    mode $channel ${mag}ooo $xx $yy $zz
 *	}
 *   }
 *
 *   # This example throws away any arguments passed to this alias
 *   /alias booya (void) { echo Booya! }
 */
static ArgList *	parse_arglist (char *arglist)
{
	char *	this_term;
	char *	next_term;
	char *	varname;
	char *	modifier, *value;
	int	arg_count = 0;
	ArgList *args = new_malloc(sizeof(ArgList));

	args->void_flag = args->dot_flag = 0;
	for (this_term = arglist; *this_term; this_term = next_term,arg_count++)
	{
		while (isspace(*this_term))
			this_term++;
		next_in_comma_list(this_term, &next_term);
		if (!(varname = next_arg(this_term, &this_term)))
			continue;
		if (!my_stricmp(varname, "void")) {
			args->void_flag = 1;
			break;
		} else if (!my_stricmp(varname, "...")) {
			args->dot_flag = 1;
			break;
		} else {
			args->vars[arg_count] = m_strdup(varname);
			args->defaults[arg_count] = NULL;

			if (!(modifier = next_arg(this_term, &this_term)))
				continue;
			if (!my_stricmp(modifier, "default"))
			{
			    if (!(value = new_next_arg(this_term, &this_term)))
				continue;
			    args->defaults[arg_count] = m_strdup(value);
			}
		}
	}
	args->vars[arg_count] = NULL;
	return args;
}

static void	destroy_arglist (ArgList *arglist)
{
	int	i = 0;

	if (!arglist)
		return;

	for (i = 0; ; i++)
	{
		if (!arglist->vars[i])
			break;
		new_free(&arglist->vars[i]);
		new_free(&arglist->defaults[i]);
	}
	new_free((char **)&arglist);
}

static void	prepare_alias_call (void *al, char **stuff)
{
	ArgList *args = (ArgList *)al;
	int	i;

	if (!args)
		return;

	for (i = 0; args->vars[i]; i++)
	{
		char	*next_val;
		char	*expanded = NULL;
		int	af;

		/* Last argument on the list and no ... argument? */
		if (!args->vars[i + 1] && !args->dot_flag && !args->void_flag)
		{
			next_val = *stuff;
			*stuff = empty_string;
		}

		/* Yank the next word from the arglist */
		else
			next_val = next_arg(*stuff, stuff);

		if (!next_val || !*next_val)
		{
			if ((next_val = args->defaults[i]))
				next_val = expanded = expand_alias(next_val, *stuff, &af, NULL);
			else
				next_val = empty_string;
		}

		/* Add the local variable */
		add_local_alias(args->vars[i], next_val, 0);
		if (expanded)
			new_free(&expanded);
	}

	/* Throw away rest of args if wanted */
	if (args->void_flag)
		*stuff = empty_string;
}

/**************************** INTERMEDIATE INTERFACE *********************/
void	add_cmd_alias	(char *name, ArgList *arglist, char *stuff, int noisy)
{
	int		af;
	Alias *		tmp = NULL;
	int		count;
	char *		save;

	save = name = remove_brackets(name, NULL, &af);
	tmp = find_cmd_alias(name, &count, 1);

	/*
	 * Then we fill in the interesting details.
	 */
	if (!tmp->filename || strcmp(tmp->filename, current_package()))
		malloc_strcpy(&tmp->filename, current_package());
	malloc_strcpy(&tmp->stuff, stuff);
	new_free(&tmp->stub);
	tmp->global = loading_global;
	tmp->arglist = arglist;

	alias_total_allocated++;
	alias_total_bytes_allocated += strlen(tmp->stuff);
	if (noisy)
		say("Alias	%s added [%s]", name, stuff);

	new_free(&name);
	return;
}

void	add_cmd_stub_alias  (char *name, char *stuff, int noisy)
{
	Alias *	tmp = NULL;
	int 	count, af;

	name = remove_brackets(name, NULL, &af);
	tmp = find_cmd_alias(name, &count, 1);

	if (!tmp->filename || strcmp(tmp->filename, current_package()))
		malloc_strcpy(&tmp->filename, current_package());
	malloc_strcpy(&(tmp->stub), stuff);
	new_free(&tmp->stuff);
	tmp->global = loading_global;

	alias_total_allocated++;
	alias_total_bytes_allocated += strlen(tmp->stub);
	if (noisy)
		say("Alias %s stubbed to file %s", name, stuff);

	new_free(&name);
	return;
}

/*
 * Fetch the Alias value associated with 'name', doing any requisite
 * name completion, or unstubbing as needed.  'unstub' must always 
 * always always be '1', unless you are prepared to handle a stubbed
 * alias entry (which is basically nobody).  Another side effect of
 * unstub == 0 is that command completion is suppressed; only an exact
 * match of 'name' is returned.  This is used only by the command stub
 * creater, who needs these semantics.  Please don't abuse this. =)
 *
 * Also, if 'exact' is 1, then it only returns exact matches.  This used
 * to be handled gracefully by just checking 'cnt', but unfortunately,
 * some callers need to be assured that unstubbing will *not* occur unless
 * a specific, exact alias name exists.
 */
/*
 * XXX Ugh.  This is only ever called by /alias or /stub -- is all this
 * browbeating with unstubbing really neccesary here?
 */
Alias *	find_cmd_alias (char *name, int *count, int create)
{
	da_item *	item;
	Alias *		a;
	char *		cname;

	item = get_cmd_item(name, count, &cname);
	/* Specifically defeat command completion if 'create' == 1. */
	if (!item || !item->alias || (*count >= 0 && create))
	{
		if (create)
		{
			a = make_new_Alias(name);
			add_alias_command_entry(name, a);
			add_alias_func_entry(name, a);
			return a;
		}
		else
			return NULL;
	}

	if (!(a = (Alias *)item->alias))
		panic("Eh, this shouldn't happen");

	if ((*count == 1 || *count < 0) && a->stub)
		a = check_cmd_alias_unstub(a, count);

	if (a && !a->stuff)
		panic("a->stuff is NULL and it shouldn't be.");

	/* Take *THAT* */
	if (!a && create)
		return find_cmd_alias(name, count, create);

	return a;
}


static Alias *	check_cmd_alias_unstub (Alias *item, int *count)
{
	char *	file;
	char *	name;

	file = LOCAL_COPY(item->stub);
	name = LOCAL_COPY(item->name);
	delete_cmd_alias(name, 0);

	item = NULL;
	if (!unstub_in_progress)
	{
		unstub_in_progress = 1;
		load("LOAD", file, empty_string);
		unstub_in_progress = 0;
	}

	/*
	 * At this opint, we have to see if 'item' was
	 * redefinied by the /load.  We call find_cmd_alias
	 * recursively to pick up the new value.
	 */
	/* 
	 * We must end up with exactly the same alias that
	 * we were starting with.  This is because it might
	 * be possible to de-stub an alias by way of command
	 * completion, but we do NOT want to return another
	 * alias that also completes, if the alias we started
	 * looking at has disappeared.  Ok?
	 */
	if ((item = find_cmd_alias(name, count, 0)))
		if (strcmp(item->name, name))
			return NULL;

	return item;
}

void	delete_cmd_alias (char *name, int noisy)
{
	Alias *	item;
	char *	name_copy;

	name_copy = LOCAL_COPY(name);
	name = name_copy;
	upper(name);

	if ((item = (Alias *)remove_alias_command_entry(name)))
	{
		remove_alias_func_entry(name);
		new_free(&item->name);
		new_free(&item->stuff);
		new_free(&item->stub);
		new_free(&item->filename);
		destroy_arglist(item->arglist);
		new_free((char **)&item);
		if (noisy)
			say("Alias	%s removed", name);
	}
	else if (noisy)
		say("No such alias: %s", name);
}

/*
 * This function is strictly O(N).  Its possible to address this.
 */
static void	list_cmd_alias (char *name)
{
        int             count;
        void *          ptr;
        Alias *         a;
        int             len;
        size_t          first_dot,
                        curr_struct_len = 0;
        const char *    curr_struct = NULL;
        const char *    s;
        const char *    script;

	if (name)
	{
		upper(name);
		len = strlen(name);
	}
	else
		len = 0;

	say("Aliases:");
	for (count = 0; (ptr = foreach_alias_item(&count)); count++)
	{
		a = (Alias *)ptr;

		if (name)
		{
			int	c = strncmp(a->name, name, len);
			if (c < 0)
				continue;
			if (c > 0)
				break;
		}

		script = *a->filename ? a->filename : "*";

                if ((s = strchr(a->name + len, '.')))
                {
                        first_dot = s - a->name;

                        if (!curr_struct ||
                            (first_dot != curr_struct_len) ||
                            strncmp(a->name, curr_struct, first_dot))
                        {
                                say("[%s]\t%*.*s\t<Structure>",
                                        script, (int)first_dot,
                                        (int)first_dot, a->name);
                                curr_struct = a->name;
                                curr_struct_len = first_dot;
                        }
                }
                else
                {
                        if (a->stub)
                                say("[%s]\t%s STUBBED TO %s",
                                        script, a->name, a->stub);
                        else
                                say("[%s]\t%s\t%s",
                                        script, a->name, a->stuff);
                }
	}
}

void	unload_cmd_alias (const char *filename)
{
        int     cnt;
        void *  ptr;
        Alias * a;

        for (cnt = 0; (ptr = foreach_alias_item(&cnt)); cnt++)
        {
                a = (Alias *)ptr;
                if (!strcmp(a->filename, filename))
                        delete_cmd_alias(a->name, 0);
        }
}

/*
 * This function is strictly O(N).  This should probably be addressed.
 */
char **	glob_cmd_alias (char *name, int *howmany)
{
        int             cnt;
        void *          ptr;
        Alias *         a;
        int             cmp;
        int             len;
        char **         matches = NULL;
        int             matches_size = 5;

        len = strlen(name);
        *howmany = 0;
        matches = RESIZE(matches, char *, matches_size);

        for (cnt = 0; (ptr = foreach_alias_item(&cnt)); cnt++)
        {
                a = (Alias *)ptr;

                if (!(cmp = strncmp(name, a->name, len)))
                {
                        if (strchr(a->name + len, '.'))
                                continue;

                        matches[*howmany] = m_strdup(a->name);
                        (*howmany)++;
                        if (*howmany == matches_size)
                        {
                                matches_size += 5;
                                RESIZE(matches, char *, matches_size);
                        }
                }
                else if (cmp < 0)
                        break;
        }

        if (*howmany)
                matches[*howmany] = NULL;
        else
                new_free((char **)&matches);

        return matches;
}

/*
 * This function is strictly O(N).  This should probably be addressed.
 */
char **	pmatch_cmd_alias (char *name, int *howmany)
{
        int             cnt;
        void *          ptr;
        Alias *         a;
        int             len;
        char **         matches = NULL;
        int             matches_size = 5;

        len = strlen(name);
        *howmany = 0;
        matches = RESIZE(matches, char *, matches_size);

        for (cnt = 0; (ptr = foreach_alias_item(&cnt)); cnt++)
        {
                a = (Alias *)ptr;

                if (wild_match(name, a->name))
                {
                        matches[*howmany] = m_strdup(a->name);
                        *howmany += 1;
                        if (*howmany == matches_size)
                        {
                                matches_size += 5;
                                RESIZE(matches, char *, matches_size);
                        }
                }
        }

        if (*howmany)
                matches[*howmany] = NULL;
        else
                new_free((char **)&matches);

        return matches;
}

void	save_aliases (FILE *fp, int do_all)
{
        int     cnt = 0;
        void *  ptr;
        Alias * a;

        for (cnt = 0; (ptr = foreach_alias_item(&cnt)); cnt++)
        {
                a = (Alias *)ptr;
                if (!a->global || do_all)
                {
                        if (a->stub)
                                fprintf(fp, "STUB ALIAS %s %s",
                                        a->name, a->stub);
                        else
                                fprintf(fp, "ALIAS %s %s\n",
                                        a->name, a->stuff);
                }
        }
}

/************************************************************************/
static char *	parse_line_alias_special (char *name, char *what, const char *args, int d1, int d2, void *arglist, int function)
{
	int	old_last_function_call_level = last_function_call_level;
	char *	result = NULL;
	char *	my_args;

	make_local_stack(name);
	my_args = LOCAL_COPY(args);
	prepare_alias_call(arglist, &my_args);
	if (function)
	{
		last_function_call_level = wind_index;
		add_local_alias("FUNCTION_RETURN", empty_string, 0);
	}

	will_catch_return_exceptions++;
	parse_line(NULL, what, args, d1, d2);
	will_catch_return_exceptions--;
	return_exception = 0;

	if (function)
	{
		result = get_variable("FUNCTION_RETURN");
		last_function_call_level = old_last_function_call_level;
	}
	destroy_local_stack();
	return result;
}

char *	parse_line_with_return (char *name, char *what, const char *args, int d1, int d2)
{
	return parse_line_alias_special(name, what, args, d1, d2, NULL, 1);
}

/*
 * call_user_function: Executes a user alias (by way of parse_command.
 * The actual function ends up being routed through execute_alias (below)
 * and we just keep track of the retval and stuff.  I dont know that anyone
 * depends on command completion with functions, so we can save a lot of
 * CPU time by just calling execute_alias() directly.
 */
char 	*call_user_function (void *alias_ptr, const char *args)
{
	Alias *	alias = (Alias *)alias_ptr;
	char *	retval = NULL;
	int 	cnt;
	char *	copy;

	if (alias->stub)
		alias = check_cmd_alias_unstub(alias, &cnt);

	if (!alias)
		return m_strdup(empty_string);

	copy = LOCAL_COPY(alias->name);
	if ((retval = parse_line_alias_special(copy, alias->stuff, 
						args, 0, 1, 
						alias->arglist, 1)))
		return retval;

	return m_strdup(empty_string);
}

/*
 * Just like call_user_function, but it throws away the return value.
 */
void	call_user_alias	(void *alias_ptr, const char *args)
{
	Alias *	alias = (Alias *)alias_ptr;
	char *	retval;
	int	cnt = 0;

	if (alias->stub)
		alias = check_cmd_alias_unstub(alias, &cnt);

	if (!alias)
		return;

	if ((retval = parse_line_alias_special(alias->name, alias->stuff, args,
						 0, 1, alias->arglist, 0)))
		new_free(&retval);
}


