/* $EPIC:$ */
/*
 * if.c: the IF, WHILE, FOREACH, DO, FE, FEC, and FOR commands for IRCII 
 *
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-1999 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "alias.h"
#include "if.h"
#include "ircaux.h"
#include "output.h"
#include "parse.h"
#include "vars.h"
#include "commands.h"
#include "window.h"


/*
 * next_expr finds the next expression delimited by brackets. The type
 * of bracket expected is passed as a parameter. Returns NULL on error.
 */
__inline__
char *	my_next_expr (char **args, char type, int whine)
{
	char	*ptr,
		*ptr2,
		*ptr3;

	if (!*args)
		return NULL;
	ptr2 = *args;
	if (!*ptr2)
		return 0;

	if (*ptr2 != type)
	{
		if (whine)
			say("Expression syntax");
		return 0;
	}

	ptr = MatchingBracket(ptr2 + 1, type, (type == '(') ? ')' : '}');
	if (!ptr)
	{
		say("Unmatched '%c'", type);
		return 0;
	}
	*ptr = '\0';

	do
		ptr2++;
	while (my_isspace(*ptr2));

	ptr3 = ptr+1;
	while (my_isspace(*ptr3))
		ptr3++;
	*args = ptr3;
	if (*ptr2)
	{
		ptr--;
		while (my_isspace(*ptr))
			*ptr-- = '\0';
	}
	return ptr2;
}

char *	next_expr_failok (char **args, char type)
{
	return my_next_expr (args, type, 0);
}

char *	next_expr (char **args, char type)
{
	return my_next_expr (args, type, 1);
}


/*
 * All new /if command -- (jfn, 1997)
 *
 * Here's the plan:
 *
 *		if (expr) ......
 *		if (expr) {......}
 *		if (expr) {......} {......}
 *		if (expr) {......} else {......}
 *		if (expr) {......} elsif (expr2) {......}
 * etc.
 */

BUILT_IN_COMMAND(ifcmd)
{
	int 	unless_cmd;
	char *	current_expr;
	char *	current_expr_val;
	int 	result;
	char *	current_line = NULL;
	int 	flag = 0;

	unless_cmd = (*command == 'U');
	if (!subargs)
		subargs = empty_string;

	if (!args)
		yell("Usage: %s (expr) {true} [ELSIF (expr) {true}] "
			"[ELSE] {false}", command);

	while (args && *args)
	{
		while (my_isspace(*args))
			args++;

		current_expr = next_expr(&args, '(');
		if (!current_expr)
		{
			error("IF: Missing expression");
			return;
		}
		current_expr_val = parse_inline(current_expr, subargs, &flag, 1);
		if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
			yell("%s expression expands to: (%s)", 
				command, current_expr_val);

		result = check_val(current_expr_val);
		new_free(&current_expr_val);

		if (*args == '{')
			current_line = next_expr(&args, '{');
		else
			current_line = args, args = NULL;

		/* If the expression was FALSE for IF, and TRUE for UNLESS */
		if (unless_cmd == result)
		{
			if (args)
			{
				if (!my_strnicmp(args, "elsif ", 6))
				{
					args += 6;
					continue;
				}
				else if (!my_strnicmp(args, "else ", 5))
					args += 5;

				while (my_isspace(*args))
					args++;

				if (*args == '{')
					current_line = next_expr(&args, '{');
				else
					current_line = args, args = NULL;

			}
			else
				current_line = NULL;
		}

		if (current_line)
			parse_line(NULL, current_line, subargs, 0, 0);

		break;
	}
}

BUILT_IN_COMMAND(docmd)
{
	char	*body,
		*expr,
		*cmd,
		*ptr,
		*newexp = NULL;
	int	args_used = 0,
		result;

	if (!subargs)
		subargs = empty_string;

	/* Old style /do .... */
	if (*args != '{')
	{
		parse_line (NULL, args, subargs, 0, 0);
		return;
	}

	/* Extract the {...} stuff */
	if (!(body = next_expr(&args, '{')))	/* } */
	{
		error("DO: unbalanced {");	/* } */
		return;
	}

	/* If there isnt a while (...) clause on the end... */
	if (!args || !*args || !(cmd = next_arg(args, &args)) ||
		my_stricmp(cmd, "while"))
	{
		parse_line(NULL, body, subargs, 0, 0);
		return;
	}


	/* Full blown /do {...} while (...) */
	if (!(expr = next_expr(&args, '(')))	/* ) */
	{
		error("DO: unbalanced (");	/* ) */
		return;
	}
	newexp = alloca(strlen(expr) + 1);

	will_catch_break_exceptions++;
	will_catch_continue_exceptions++;
	while (1)
	{
		parse_line(NULL, body, subargs ? subargs : empty_string, 0, 0);

		if (break_exception)
		{
			break_exception = 0;
			break;
		}
		if (continue_exception)
		{
			continue_exception = 0;
			continue;
		}
		if (return_exception)
			break;

		strcpy(newexp, expr);
		ptr = parse_inline(newexp, subargs, &args_used, 1);
		result = check_val(ptr);
		new_free(&ptr);

		if (!result)
			break;
	}
	will_catch_break_exceptions--;
	will_catch_continue_exceptions--;

	return;
}

BUILT_IN_COMMAND(whilecmd)
{
	char	*exp = NULL,
		*ptr,
		*body = NULL,
		*newexp = NULL;
	int	a;			/* Dummy variable */
	int 	whileval = !strcmp(command, "WHILE");

	if (!subargs)
		subargs = empty_string;

	if (!(ptr = next_expr(&args, '(')))		/* ) */
	{
		error("WHILE: missing boolean expression");
		return;
	}
	exp = LOCAL_COPY(ptr);
	newexp = alloca(strlen(exp) + 1);

	if (!(ptr = next_expr_failok(&args, '{')))	/* } */
		ptr = args;
	body = LOCAL_COPY(ptr);

	will_catch_break_exceptions++;
	will_catch_continue_exceptions++;
	for (;;)
	{
		strcpy(newexp, exp);
		ptr = parse_inline(newexp, subargs, &a, 1);
		if (check_val(ptr) != whileval)
			break;
		new_free(&ptr);

		parse_line(NULL, body, subargs, 0, 0);
		if (continue_exception)
		{
			continue_exception = 0;
			continue;
		}
		if (break_exception)
		{
			break_exception = 0;
			break;
		}
		if (return_exception)
			break;
	}
	will_catch_break_exceptions--;
	will_catch_continue_exceptions--;
	new_free(&ptr);
}

BUILT_IN_COMMAND(foreach)
{
	char	*struc = NULL,
		*mstruc = NULL,
		*ptr,
		*body = NULL,
		*var = NULL;
	char	**sublist;
	int	total;
	int	i;
	int	slen;
	int	list = VAR_ALIAS;
	int	af;

	if (!subargs)
		subargs = empty_string;

	while (args && my_isspace(*args))
		args++;

	if (*args == '-')
		args++, list = COMMAND_ALIAS;

	if (!(ptr = new_next_arg(args, &args)))
	{
		error("FOREACH: missing structure expression");
		return;
	}

	if (!(var = next_arg(args, &args)))
	{
		error("FOREACH: missing variable");
		return;
	}

	while (*args && *args != '{')		/* } */
		args++;

	if (!(body = next_expr(&args, '{')))
	{
		error("FOREACH: missing statement");
		return;
	}

	mstruc = remove_brackets(ptr, subargs, &af);
	struc = LOCAL_COPY(mstruc);
	new_free(&mstruc);

	if ((sublist = get_subarray_elements(struc, &total, list)) == NULL)
		return;		/* Nothing there. */

	slen = strlen(struc);

	/* Hrm. should it catch breaks and continues? */
	will_catch_break_exceptions++;
	will_catch_continue_exceptions++;
	for (i = 0; i < total; i++)
	{
		add_local_alias(var, sublist[i] + slen + 1, 0);

		parse_line(NULL, body, subargs, 0, 0);

		if (continue_exception)
		{
			continue_exception = 0;
			continue;
		}
		if (break_exception)
		{
			break_exception = 0;
			break;
		}
		if (return_exception)
			break;
	}
	will_catch_break_exceptions--;
	will_catch_continue_exceptions--;

	for (i = 0; i < total; i++)
		new_free(&sublist[i]);
	new_free((char **)&sublist);
}

/*
 * FE:  Written by Jeremy Nelson (jnelson@acronet.net)
 *
 * FE: replaces recursion
 * XXX Rewrite me please! =)
 */
BUILT_IN_COMMAND(fe)
{
	int	argsused;
	int	fe_command = !strcmp(command, "FE");
	char *	raw_word_list = NULL;
	char *	e_word_list = NULL;
	char *	word_list;
	char *	raw_var_list;
	char **	var_list;
	int	num_vars;
	char *	commands;
	char *	word = NULL;		/* XXX Bogus */
	int	x;

	if (!subargs)
		subargs = empty_string;

	/*
	 * Extract the word list
	 */
	if (!(raw_word_list = next_expr(&args, '(')))		/* ) */
	{
		error("%s: Missing word list", command);
		return;
	}

	e_word_list = expand_alias(raw_word_list, subargs, &argsused, NULL);
	word_list = LOCAL_COPY(e_word_list);
	new_free(&e_word_list);

	/*
	 * Extract the variable list
	 */
	raw_var_list = args;
	if (!(args = strchr(args, '{')))	
	{
		error("%s: Missing commands", command);
		return;
	}
	*args++ = 0;

	/*
	 * Parse out all the lvals in the variable list
	 */
	if ((num_vars = splitw(raw_var_list, &var_list)) == 0)
	{
		error("%s: You did not specify any variables", command);
		new_free(&var_list);
		return;
	}


	/*
	 * Extract the block-command
	 */
	*--args = '{';		/* XXX Bleh */
	if (!(commands = next_expr(&args, '{')))
	{
		error("%s: Missing }", command);
		new_free(&var_list);
		return;
	}

	/*
	 * Is this an fec command?  Then prepare the 'word' stuff.
	 */
	if (fe_command == 0)
	{ 
		word = alloca(2);
		word[1] = 0;
	}

	/*
	 * Iterate over the entire word list, grabbing words a few at
	 * a time until they are exhausted.
	 */
	will_catch_break_exceptions++;
	will_catch_continue_exceptions++;
	while (word_list && *word_list)
	{
		for (x = 0; x < num_vars; x++)
		{
			if (fe_command != 0)
			{
			    if (!(word = new_next_arg(word_list, &word_list)))
				break;
			}
			else
			{
				if (!(*word = *word_list++))
				{
					word_list = NULL;	/* XXX */
					break;
				}
			}

			add_local_alias(var_list[x], word, 0);
		}
		for (; x < num_vars; x++)
			add_local_alias(var_list[x], empty_string, 0);

		parse_line(NULL, commands, subargs, 0, 0);

		if (continue_exception)
		{
			continue_exception = 0;
			continue;
		}
		if (break_exception)
		{
			break_exception = 0;
			break;
		}
		if (return_exception)
			break;
	}
	will_catch_break_exceptions--;
	will_catch_continue_exceptions--;
	new_free(&var_list);
}

/* 
 * FOR command..... prototype: 
 *  for (commence,evaluation,iteration)
 * in the same style of C's for, the for loop is just a specific
 * type of WHILE loop.
 *
 * IMPORTANT: Since ircII uses ; as a delimeter between commands,
 * commas were chosen to be the delimiter between expressions,
 * so that semicolons may be used in the expressions (think of this
 * as the reverse as C, where commas seperate commands in expressions,
 * and semicolons end expressions.
 */
BUILT_IN_COMMAND(forcmd)
{
	char        *working        = (char *) 0;
	char        *commence       = (char *) 0;
	char        *evaluation     = (char *) 0;
	char        *lameeval       = (char *) 0;
	char        *iteration      = (char *) 0;
	char        *blah           = (char *) 0;
	char        *commands       = (char *) 0;
	int         argsused        = 0;

	if (!subargs)
		subargs = empty_string;

	/* Get the whole () thing */
	if (!(working = next_expr(&args, '(')))
	{
		error("FOR: missing closing parenthesis");
		return;
	}
	commence = LOCAL_COPY(working);

	/* Find the beginning of the second expression */
	if (!(evaluation = strchr(commence, ',')))
	{
		error("FOR: no components!");
		return;
	}
	do 
		*evaluation++ = 0;
	while (my_isspace(*evaluation));
	lameeval = alloca(strlen(evaluation) + 1);

	/* Find the beginning of the third expression */
	if (!(iteration = strchr(evaluation, ',')))
	{
		error("FOR: Only two components!");
		return;
	}
	do 
		*iteration++ = 0;
	while (my_isspace(*iteration));

	working = args;
	while (my_isspace(*working))
		*working++ = 0;

	if (!(working = next_expr(&working, '{')))		/* } */
	{
		error("FOR: badly formed commands");
		return;
	}
	commands = LOCAL_COPY(working);

	parse_line(NULL, commence, subargs, 0, 0);

	will_catch_break_exceptions++;
	will_catch_continue_exceptions++;
	for (;;)
	{
		strcpy(lameeval, evaluation);
		blah = parse_inline(lameeval, subargs, &argsused, 1);
		if (!check_val(blah))
			break;

		new_free(&blah);
		parse_line(NULL, commands, subargs, 0, 0);

		if (break_exception)
		{
			break_exception = 0;
			break;
		}
		if (continue_exception)
			continue_exception = 0;	/* Dont continue here! */
		if (return_exception)
			break;

		parse_line(NULL, iteration, subargs, 0, 0);
	}
	will_catch_break_exceptions--;
	will_catch_continue_exceptions--;

	new_free(&blah);
}

/*

  Need to support something like this:

	switch (text to be matched)
	{
		(sample text)
		{
			...
		}
		(sample text2)
		(sample text3)
		{
			...
		}
		...
	}

How it works:

	The command is technically made up a single (...) expression and
	a single {...} expression.  The (...) expression is taken to be
	regular expando-text (much like the (...) body of /fe.

	The {...} body is taken to be a series of [(...)] {...} pairs.
	The [(...)] expression is taken to be one or more consecutive
	(...) structures, which are taken to be text expressions to match
	against the header text.  If any of the (...) expressions are found
	to match, then the commands in the {...} body are executed.

	There may be as many such [(...)] {...} pairs as you need.  However,
	only the *first* pair found to be matching is actually executed,
	and the others are ignored, so placement of your switches are
	rather important:  Put your most general ones last.

*/
BUILT_IN_COMMAND(switchcmd)
{
	char 	*control, 
		*body, 
		*header, 
		*commands;
	int 	af;

	if (!subargs)
		subargs = empty_string;

	if (!(control = next_expr(&args, '(')))
	{
		say("SWITCH: String to be matched not found where expected");
		return;
	}

	control = expand_alias(control, subargs, &af, NULL);
	if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
		yell("%s expression expands to: (%s)", command, control);

	if (!(body = next_expr(&args, '{')))
		say("SWITCH: Execution body not found where expected");

	while (body && *body)
	{
		int hooked = 0;

		while (*body == '(')
		{
			if (!(header = next_expr(&body, '(')))
			{
				say("SWITCH: Case label not found where expected");
				new_free(&control);
				return;
			}
			header = expand_alias(header, subargs, &af, NULL);
			if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
				yell("%s expression expands to: (%s)", 
					command, header);

			if (wild_match(header, control))
				hooked = 1;
			new_free(&header);
			if (*body == ';')
				body++;		/* ugh. */
		}

		if (!(commands = next_expr(&body, '{')))
		{
			say("SWITCH: case body not found where expected");
			break;
		}

		if (hooked)
		{
			will_catch_break_exceptions++;
			parse_line(NULL, commands, subargs, 0, 0);
			if (break_exception)
				break_exception = 0;
			will_catch_break_exceptions--;
			break;
		}

		if (*body == ';')
			body++;		/* grumble */
	}

	new_free(&control);
}

BUILT_IN_COMMAND(repeatcmd)
{
	char *	num_expr = NULL;
	int 	value;

	if (!subargs)
		subargs = empty_string;

	while (isspace(*args))
		args++;

	if (*args == '(')
	{
		char *		tmp_val;
		char *		dumb_copy;
		int 		argsused;

		num_expr = next_expr(&args, '(');
		dumb_copy = LOCAL_COPY(num_expr);

		tmp_val = parse_inline(dumb_copy, subargs, &argsused, 1);
		value = my_atol(tmp_val);
		new_free(&tmp_val);
	}
	else
	{
		char *		tmp_val;
		int 		af;

		num_expr = new_next_arg(args, &args);
		tmp_val = expand_alias(num_expr, subargs, &af, NULL);
		value = my_atol(tmp_val);
		new_free(&tmp_val);
	}

	if (value <= 0)
		return;

	/* Probably want to catch break and continue here */
	while (value--)
		parse_line(NULL, args, subargs, 0, 0);

	return;
}
