/* $EPIC:$ */
/*
 * network.c -- handles stuff dealing with connecting and name resolving
 *
 * Copyright  1995-2000 Jeremy Nelson
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "ircaux.h"
#include "vars.h"
#include "network.h"
#include "newio2.h"

#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif


/*
	#define SERVICE_SERVER 0
	#define SERVICE_CLIENT 1
	#define PROTOCOL_TCP 0
	#define PROTOCOL_UDP 1
*/
/*
 * connect_by_number:  Wheeeee. Yet another monster function i get to fix
 * for the sake of it being inadequate for extension.
 *
 * we now take five arguments:
 *
 *	- hostname - name of the host (pathname) to connect to (if applicable)
 *	- port    - port number to connect to or listen on (0 if you dont care)
 *		    port must be in NETWORK ORDER! (this is a change)
 *	- service -	0 - set up a listening socket
 *			1 - set up a connecting socket
 *	- protocol - 	0 - use the TCP protocol
 *			1 - use the UDP protocol
 *	- callback -	For a connecting socket, this function will be
 *			called when the connect() fails or succeeds.
 *			This implies nonblocking connects, of course.
 *
 * Returns:
 *	Non-negative number -- new file descriptor ready for use
 *	-1 -- could not open a new file descriptor or 
 *		an invalid value for the protocol was specified
 *	-2 -- call to bind() failed
 *	-3 -- call to listen() failed.
 *	-4 -- call to connect() failed
 *	-5 -- call to getsockname() failed
 *	-6 -- the name of the host could not be resolved
 *	-7 -- invalid or unsupported request
 *
 *
 * Credit: I couldnt have put this together without the help of 4.4BSD
 * User Supplimentary Document #20 (Inter-process Communications tutorial)
 */
int 	connect_by_number (const char *hostn, unsigned short *port, int service, int protocol, int (*callback) (int, int, char *))
{
	int 	fd = -1;
	int 	is_unix = (hostn && *hostn == '/');
	int 	sock_type, proto_type;

	sock_type = (is_unix) ? AF_UNIX : AF_INET;
	proto_type = (protocol == PROTOCOL_TCP) ? SOCK_STREAM : SOCK_DGRAM;

	if ((fd = socket(sock_type, proto_type, 0)) < 0)
		return -1;

	set_socket_options (fd);

	/* Unix domain server */
#ifdef HAVE_SYS_UN_H
	if (is_unix)
	{
		struct sockaddr_un 	name;

		my_bzero(name);
		name.sun_family = AF_UNIX;
		strlcpy(name.sun_path, hostn, sizeof(name.sun_path));
#ifdef HAVE_SUN_LEN
# ifdef SUN_LEN
		name.sun_len = SUN_LEN(&name);
# else
		name.sun_len = strlen(name.sun_path) + 1;
# endif
#endif

		if (is_unix && (service == SERVICE_SERVER))
		{
			if (bind(fd, (SA *)&name, strlen(name.sun_path) + 2))
				return close(fd), -2;

			if (protocol == PROTOCOL_TCP)
				if (listen(fd, 4) < 0)
					return close(fd), -3;
		}

		/* Unix domain client */
		else if (service == SERVICE_CLIENT)
		{
			/* Nonblocking connect? */
			if (callback)
			{
				set_non_blocking(fd);
				connect(fd, (SA *)&name, 
					strlen(name.sun_path) + 2);
				set_blocking(fd);
				nonblocking_connect_callback(fd, callback);
				return fd;
			}

			/* Blocking connect */
			else
			{
			    alarm(get_int_var(CONNECT_TIMEOUT_VAR));
			    if (connect(fd, (SA *)&name, 
					strlen(name.sun_path) + 2) < 0)
			    {
				alarm(0);
				return close(fd), -4;
			    }
			    alarm(0);
			}
		}
	}
	else
#endif

	/* Inet domain server */
	if (!is_unix && (service == SERVICE_SERVER))
	{
		ISA	name;
		int 	length;

		my_bzero(name);
		name.sin_family = AF_INET;
		name.sin_addr.s_addr = htonl(INADDR_ANY);
		name.sin_port = *port;

#ifdef IP_PORTRANGE
		if (getenv("EPIC_USE_HIGHPORTS"))
		{
			int	ports = IP_PORTRANGE_HIGH;
			setsockopt(fd, IPPROTO_IP, IP_PORTRANGE, 
					(char *)&ports, sizeof(ports));
		}
#endif

		if (bind(fd, (SA *)&name, sizeof(name)))
			return close(fd), -2;

		length = sizeof(name);
		if (getsockname(fd, (SA *)&name, &length))
			return close(fd), -5;
		*port = name.sin_port;

		if (protocol == PROTOCOL_TCP)
			if (listen(fd, 4) < 0)
				return close(fd), -3;
	}

	/* Inet domain client */
	else if (!is_unix && (service == SERVICE_CLIENT))
	{
		ISA	server;

		/*
		 * Doing this bind is bad news unless you are sure that
		 * the hostname is valid.  This is not true for me at home,
		 * since i dynamic-ip it.
		 */
		if (LocalHostName)
		{
			ISA	localaddr;

			my_bzero(localaddr);
			localaddr.sin_family = AF_INET;
			localaddr.sin_addr = LocalHostAddr;
			localaddr.sin_port = 0;
			if (bind(fd, (SA *)&localaddr, sizeof(localaddr)))
				return close(fd), -2;
		}

		my_bzero(server);
		server.sin_addr = inet_anyton(hostn);
		if (SOCKADDR(server) == INADDR_NONE)
		{
			errno = -1;
			return close(fd), -6;
		}
		server.sin_family = AF_INET;
		server.sin_port = *port;

		/* Nonblocking connect? */
		if (callback)
		{
			set_non_blocking(fd);
			connect(fd, (SA *)&server, sizeof(server));
			set_blocking(fd);
			nonblocking_connect_callback(fd, callback);
			return fd;
		}

		/* Blocking connect */
		else
		{
			alarm(get_int_var(CONNECT_TIMEOUT_VAR));
			if (connect (fd, (SA *)&server, sizeof(server)) < 0)
			{
				alarm(0);
				return close(fd), -4;
			}
			alarm(0);
		}
	}

	/* error */
	else
		return close(fd), -7;

	return fd;
}


/******************** HOSTNAME LOOKUP FUNCTIONS **************************/
/*
 * Input:  An internet address (struct in_addr)
 * Output: A hostname entry for that Internet Address
 */
struct hostent *	inet_ntoh (struct in_addr addr)
{
	struct hostent *host = NULL;

	if (ADDRESS(addr) != INADDR_NONE)
	{
		alarm(1);
		host = gethostbyaddr((char *)&addr, 4, AF_INET);
		alarm(0);
	}
	return host;
}

/*
 * Input:  A hostname entry
 * Output: An internet address (struct in_addr) for that host
 */
struct in_addr *	inet_hton (const struct hostent *host)
{
	if (host)
		return (IA *)host->h_addr;
	else
		return NULL;
}

/*
 * Input:  An internet hostname (eg, "ftp.microsoft.com")
 * Output: A hostname entry for that host.
 */
struct hostent *	inet_hntoh (const char *hn)
{
	struct hostent *host = NULL;

	if (hn)
	{
		alarm(1);
		host = gethostbyname(hn);
		alarm(0);
	}
	return host;
}

/*
 * Input:  A hostname entry
 * Output: The canonical internet hostname (eg, "ftp.microsoft.com")
 *	   for that host.
 */
const char *	inet_htohn (const struct hostent *host)
{
	static	char	ip[128];

	*ip = 0;
	if (host)
		strcpy(ip, host->h_name);

	return ip;
}


/*
 * Input:  A dotted-quad IP address (eg, "127.0.0.1")
 * Output: A hostname entry for that IP address.
 */
struct hostent *	inet_atoh (const char *ip)
{
	struct hostent *host = NULL;
	IA		addr;

	if (inet_aton(ip, &addr))
		host = inet_ntoh(addr);

	return host;
}

/*
 * Input:  A hostname entry
 * Output: A dotted-quad IP Address (eg, "127.0.0.1") for that host.
 */
const char *	inet_htoa (const struct hostent *host)
{
	static char	ip[128];

	*ip = 0;
	if (host)
		strcpy(ip, inet_ntoa(*(inet_hton(host))));

	return ip;
}

/*
 * Input:  Either an internet hostname (eg, "ftp.microsoft.com") or 
 *	   a dotted-quad IP address (eg, "127.0.0.1")
 * Output: A hostname entry for that host.
 */
struct hostent *	inet_anytoh (const char *stuff)
{
	struct hostent *hep = NULL;

	if (stuff && *stuff)
	{
		if (isdigit(last_char(stuff)))
			hep = inet_atoh(stuff);
		else
			hep = inet_hntoh(stuff);
	}

	return hep;
}


/*
 * Input:  An internet hostname (eg, "ftp.microsoft.com")
 * Output: The dotted-quad IP address (eg, "127.0.0.1") for that host.
 */
const char *	inet_hntoa (const char *hn)
{
	struct hostent *host;
	static char	ip [256];

	*ip = 0;
	if (hn && (host = inet_hntoh(hn)))
		strcpy(ip, inet_htoa(host));

	return ip;
}

/*
 * Input:  A dotted-quad IP address (eg, "127.0.0.1")
 * Output: The internet hostname (eg, "ftp.microsoft.com") for that host.
 */
const char *	inet_atohn (const char *ip)
{
	struct hostent *host;
	static char 	hn [256];

	*hn = 0;
	if (ip && (host = inet_atoh(ip)))
		strcpy(hn, inet_htohn(host));

	return hn;
}

/*
 * Input:  A "struct in_addr" object
 * Output: An internet hostname (eg, "ftp.microsoft.com")
 */
const char *	inet_ntohn (IA ip)
{
	static char 	host[128];
	Host *		hep;

	*host = 0;
	if ((hep = inet_ntoh(ip)))
		strcpy(host, hep->h_name);
	return host;
}

struct in_addr *	inet_hnton (const char *host)
{
	static IA	addr;
	Host *		hep;

	ADDRESS(addr) = INADDR_NONE;
	if ((hep = inet_hntoh(host)))
		addr = *inet_hton(hep);
	return &addr;
}

/*
 * Input:  A "struct in_addr" object
 * Output: Either the internet hostname (eg, "ftp.microsoft.com") if available,
 *	   otherwise the dotted-quad IP address (eg, "127.0.0.1")
 */
const char *	inet_ntoany (IA ip)
{
	static	char	host[128];
	Host *		hep;

	if ((hep = inet_ntoh(ip)))
		strcpy(host, hep->h_name);
	else
		strcpy(host, inet_ntoa(ip));
	return host;
}

/*
 * Input:  Either an internet hostname (eg, "ftp.microsoft.com" or
 *	   a dotted-quad IP address (eg, "127.0.0.1")
 * Output: A "struct in_addr" object
 */
struct in_addr	inet_anyton (const char *what)
{
	struct in_addr	addr;

	if (isdigit(last_char(what)))
		inet_aton(what, &addr);
	else 
		addr = *inet_hnton(what);

	return addr;
}

/*
 * Input:  Either an internet hostname (eg, "ftp.microsoft.com") or
 *	   a dotted-quad IP address (eg, "127.0.0.1")
 * Output: Either an internet hostname or dotted-quad IP address; whichever
 *	   was not provided as input.
 */
const char *	inet_anytoother (const char *what)
{
	/* if the last char is a digit, it's an ip, else a hostname. */
	if (isdigit(last_char(what)))
		return inet_atohn(what);
	else
		return inet_hntoa(what);
}



/************************** SOCKET OPERATIONS *******************************/
int	set_non_blocking (int fd)
{
	int	flag, rval;

#if defined(O_NONBLOCK)
	flag = O_NONBLOCK;
#elif defined(O_NDELAY)
	flag = O_NDELAY;
#elif defined(FIONBIO)
	flag = 1;
#else
	yell("Sorry!  Can't set nonblocking on this system!");
#endif

#if defined(O_NONBLOCK) || defined(O_NDELAY)
	if ((rval = fcntl(fd, F_GETFL, 0)) == -1)
		return -1;
	if (fcntl(fd, F_SETFL, rval | flag) == -1)
		return -1;
#else
	if (ioctl(fd, FIONBIO, &flag) < 0)
		return -1;
#endif
	return 0;
}

int	set_blocking (int fd)
{
	int	flag, rval;

#if defined(O_NONBLOCK)
	flag = O_NONBLOCK;
#elif defined(O_NDELAY)
	flag = O_NDELAY;
#elif defined(FIONBIO)
	flag = 0;
#else
	yell("Sorry!  Can't set nonblocking on this system!");
#endif

#if defined(O_NONBLOCK) || defined(O_NDELAY)
	if ((rval = fcntl(fd, F_GETFL, 0)) == -1)
		return -1;
	if (fcntl(fd, F_SETFL, rval & ~flag) == -1)
		return -1;
#else
	if (ioctl(fd, FIONBIO, &flag) < 0)
		return -1;
#endif
	return 0;
}


/*
 * It is possible for a race condition to exist; such that select()
 * indicates that a listen()ing socket is able to recieve a new connection
 * and that a later accept() call will still block because the connection
 * has been closed in the interim.  This wrapper for accept() attempts to
 * defeat this by making the accept() call nonblocking.
 */
int	my_accept (int s, struct sockaddr *addr, int *addrlen)
{
	int	retval;

	set_non_blocking(s);
	retval = accept(s, addr, addrlen);
	set_blocking(s);
	return retval;
}

/* set's socket options */
void 	set_socket_options (int s)
{
	int	opt;
	int	optlen = sizeof(opt);
#ifndef NO_STRUCT_LINGER
	struct linger	lin;

	lin.l_onoff = lin.l_linger = 0;
	setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&lin, optlen);
#endif

	opt = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, optlen);
	opt = 1;
	setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *)&opt, optlen);

}
