/* $EPIC:$ */
/*
 * newio2.c: A whole new way to do I/O for ircII.
 *
 * IMPORTANT!  If you register your file descriptors with these functions,
 * then you must not use read() or write() or anything else to directly access
 * the file descriptor in the way it was registered -- you must only use the 
 * functionality when the callback is asserted.  Failure to do this will
 * result in erratic and unpredictable behavior.
 *
 * Copyright  1999 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

/*
 * Here's the plan:
 *
 * We want to devise a passive callback-driven mechanism for I/O handling
 * in epic.  The old way (newio.c) was active-driven, where the client would
 * wait for "something" to happen, then call 4 different callbacks and
 * each of them would see if that "something" was something that they could
 * deal with.  The problem with that was adding a new callback was a pain and
 * it was also hardcoded into the looping mechanism, so the callbacks could 
 * not just be called on an as-needed basis.
 *
 * This new I/O model works on the premise that operations on all file
 * descriptors take place in four distinct stages:
 *
 *	Stage 1	(Event driver)		A descriptor becomes ready for 
 *					an operation
 *
 *	Stage 2 (I/O driver)		The operation is performed on
 *					the descriptor and state information
 *					is modified appropriately.
 *
 *	Stage 3 (Logical driver)	The data is broken into logical
 *					units and prepared for consumption.
 *
 *	Stage 4 (Application driver)	A single logical unit of data is
 *					processed by the application.
 *
 *
 * Control from one level to another is handled through "disciplines", 
 * which is a set of callback functions for interesting events:
 *
 *		struct	discipline_struct {
 *			read_callback		reader;
 *			write_callback		writer;
 *			eof_callblack		eof;
 *			error_callback		error;
 *		};
 * 
 * For each of these functions that are not NULL, the appropriate action is
 * watched for and handled.  When an event occurs, the function is called.
 *
 *	reader -- When a read operation is to be performed on a descriptor,
 *		  this function does the appropriate action.  At stage 2,
 *		  this function actually performs the read(2) and places the
 *		  contents into a "dirty" buffer.  At stage 3, this function
 *		  removes one logical unit from a "dirty" buffer (possibly
 *		  yielding a "clean" buffer), and at stage 4, this function
 *		  performs the appropriate action on the new logical input.
 *
 *	writer -- When a write operation is to be performed on a descriptor,
 *		  this function does the appropriate action.  At stage 2,
 *		  this function actually performs a write(2) of a "dirty"
 *		  buffer (possibly yielding a "clean" buffer).  When the 
 *		  buffer is fully exhausted, the stage 3 function is invoked
 *		  to replenish the buffer.  The stage 4 function is called
 *		  to actually provide the data, which the stage 3 function
 *		  places into the now "dirty" buffer.
 *
 *	eof --    When a read or write operation on a descriptor fails due
 *		  to an End Of File condition, this function is called.
 *
 *	error --  When a read or write operation on a descriptor fails due
 *		  to a system error or due to an unrecoverable software
 *		  condition, this function is called.
 *
 * As much as is possible, the details of each stage is hidden from others.
 *
 * Stage 1 exposes several functions:
 *	void	do_events		(int mtimeout);
 *	void	event__track 		(int fd, Discipline *d);
 *	void	event__untrack 		(int fd);
 *
 * Stage 2 exposes several functions:
 *	int	buffer__data_waiting 	(void);
 *	int	io__read 		(int fd, int unused, char *ignored);
 *	int	io__accept 		(int fd, int unused, char *ignored);
 * 	int	io__write 		(int fd, int unused, char *ignored);
 *	int	io__connect 		(int fd, int unused, char *ignored);
 *
 * Stage 3 exposes several functions:
 *	int	logical__handle_data 	(void);
 * 	void	logical__track 		(int fd, Discipline *, Discipline *);
 *	void	logical__untrack 	(int fd);
 *	int	logical__eof 		(int fd);
 *	int	logical__error 		(int fd);
 *	int	new_close 		(int fd);
 *
 * Stage 4 exposes several functions:
 *	int	no_buffering 			(int fd, Discipline *appl);
 *	int	no_buffering_readonly 		(int fd, Discipline *appl);
 *	int	exec_buffering 			(int fd, Discipline *appl);
 *	int	line_buffering 			(int fd, Discipline *appl);
 *	int	nonblocking_connect_callback 	(int fd, Writer writer);
 *	int	accept_callback 		(int fd, Reader reader);
 *
 * The following data items are interesting:
 *
 * 	io__discipline		(Stage 1 to Stage 2 i/o callbacks)
 *	logical__discipline	(Stage 2 to Stage 3 event callbacks)
 *	application_discipline	(Stage 3 to Stage 4 event callbacks)
 * Plus, the function 'logical__handle_data' transfers control from 
 * Stage 1 to Stage 2
 *
 * ***
 * General Flow of data:
 *	
 *	1) Application calls do_events
 *	2) do_events notices a descriptor ready for operation.
 *	3) If operation is a read operation
 *		4) io__discipline[fd]->read()  is called
 *		5) This function does io reading action.
 *		6) If it succeeds, places data into the fd's read buffer, 
 *			marks it as dirty
 *		   Else if it is an EOF condition, calls the 
 *			logical__discipline[fd]->eof() function.
 *		   Else if it is an error condition, calls the
 *			logical__discipline[fd]->error() function.
 *	3) Else if operation is a write operation
 *		4) io__discipline[fd]->write() is called
 *		5) This function does io writing action
 *		6) If fd's write buffer is empty, marks it as clean
 *		   Else if it is an EOF condition, calls the
 *			logical__discipline[fd]->eof() function.
 *		   Else if it is an error condition, calls the
 *			logical__discipline[fd]->error() function.
 *      7) If a socket's read buffer is dirty, then call
 *	   logical__discipline[fd]->read() function.
 *		8) Remove the first logical unit from the buffer
 *		9) If there is another full logical unit, mark the buffer
 *		   as dirty, else mark it as clean.
 *		10) Call the application__discipline[fd]->read() function.
 *	7) Else if a socket's write buffer is clean, then call
 *	   logical__discipline[fd]->write() function
 *		8) Call the application_discipline[fd]->write() function.
 *		9) Place the resulting data into the write buffer, mark the
 *		   write buffer as dirty.
 *	11) Return to application
 */
 
#include "irc.h"
#include "ircaux.h"
#include "newio2.h"
#include "output.h"
#include "input.h"
#include "network.h"

#include <sys/ioctl.h>
#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>		/* XXXX nonportable */
#endif
#ifdef HAVE_SYS_STROPTS_H
#include <sys/stropts.h>	/* XXXX nonportable */
#endif

#if defined(HAVE_SYSCONF) && defined(_SC_OPEN_MAX)
# define IO_ARRAYLEN	sysconf(_SC_OPEN_MAX)
#elif defined(FD_SETSIZE)
# define IO_ARRAYLEN	FD_SETSIZE
#else
# define IO_ARRAYLEN	NFDBITS
#endif

#define IOBUFSIZ 256*128	/* XXX - Robohak says this isnt big enough. */
		/* 
		 * XXXX This is being filled before it can empty,
		 * I havent debugged it yet, increasing the buffer
		 * size seems to solve the problem temporarily
		 */

	int	io_errno;


/************************************************************************/
typedef struct	buffer_u 
{
		char 		buffer[IOBUFSIZ];
		char *		read_pos;
		char *		write_pos;
		int		dirty;
} Buffer;


/*************************  DATA TYPES  ************************************/
	int			global_max_fd = -1;
	int			IO_ARRAYSIZE = -1;

static	Discipline **		io__discipline = NULL;

static	Buffer **		logical__readbuf = NULL;
static	Buffer **		logical__writebuf = NULL;
static	Discipline **		logical__discipline = NULL;

static	Discipline **		application__discipline = NULL;


/*************************** FUNCTIONS **********************************/
static	int	buffer__data_waiting (void);
static	int	logical__handle_data (void);
static	void	logical__track (int fd, Discipline *io, Discipline *buff, Discipline *appl);
static	void	logical__untrack (int fd);
static 	int	logical__eof (int fd);
static 	int	logical__error (int fd);


/*
 * These macros make it easier for the level 2 handlers to manipulate the
 * file descriptor data in a general way and not have to know exactly how
 * the data is actually stored and implemented.  Great time saver.
 */
#define IOX(x)		io__discipline[x]
#define IO_READ(x)	IOX(x)->reader
#define IO_WRITE(x)	IOX(x)->writer
#define IO_EOF(x)	IOX(x)->eof
#define IO_ERROR(x)	IOX(x)->error

#define READX(x)	logical__readbuf[x]
#define RBUF_BUFFER(x) 	READX(x)->buffer
#define RBUF_DIRTY(x)	READX(x)->dirty
#define RBUF_START(x)	READX(x)->read_pos
#define RBUF_END(x)	READX(x)->write_pos
#define RBUF_SIZE(x)	IOBUFSIZ

#define WRITEX(x) 	logical__writebuf[x]
#define WBUF_BUFFER(x)	WRITEX(x)->buffer
#define WBUF_DIRTY(x)	WRITEX(x)->dirty
#define WBUF_START(x)	WRITEX(x)->read_pos
#define WBUF_END(x)	WRITEX(x)->write_pos
#define WBUF_SIZE(x)	IOBUFSIZ

#define LOGICX(x)	logical__discipline[x]
#define LOGIC_READ(x)	LOGICX(x)->reader
#define LOGIC_WRITE(x)	LOGICX(x)->writer
#define LOGIC_EOF(x)	LOGICX(x)->eof
#define LOGIC_ERROR(x)	LOGICX(x)->error

#define APPLX(x)	application__discipline[x]
#define APPL_READER(x)	APPLX(x)->reader
#define APPL_WRITER(x)	APPLX(x)->writer
#define APPL_EOF(x)	APPLX(x)->eof
#define APPL_ERROR(x)	APPLX(x)->error





/************************ EVENT DRIVER ****************************/
/*
 * This is the main event driver for all of EPIC.  The specified argument is
 * the minimum number of milliseconds that must pass before this function is 
 * permitted to return when there is otherwise no activity.  In general, the
 * only function that has any business calling this function is io().  
 */
void	do_events	(int mtimeout)
{
	if (x_debug & DEBUG_NEWIO)
		yell("do_events: called");

	/*
	 * First thing we check is to see if any data activity can be done
	 * right away.  If any read buffers need to be drained or if any
	 * write buffers need to be refilled, then the callbacks to handle
	 * those actions can take an arbitrary amount of time; therefore we
	 * return right away and hope for the best.
	 */
	if (!buffer__data_waiting())
	{
		Timeval		ntimeout;
		Timeval *	timeout = &ntimeout;
		fd_set		rd, wd;

		if (x_debug & DEBUG_NEWIO)
			yell("do_events: Waiting on a select() call...");

		/*
		 * select() modifies the values passed to it, so we make our
		 * own private copies.  'readables' and 'writeables' are the
		 * master copies of the file descriptors that we are watching.
		 */
		rd = readables;
		wd = writeables;

		/*
		 * Compute the timeout.  -1 means "wait forever".
		 */
		if (mtimeout == -1)
			timeout = NULL;
		else
		{
			timeout->tv_sec = mtimeout / 1000;
			timeout->tv_usec = (mtimeout % 1000) * 1000;
		}

		/*
		 * Now we wait for something to happen.
		 */
		switch (select(global_max_fd + 1, &rd, &wd, NULL, timeout))
		{
			/* An error occured */
			case -1:
			{
				/* A SIGINT caused EINTR?  User pressed ^C */
				if (cntl_c_hit)
				{
					edit_char(3);
					cntl_c_hit = 0;
				}
				/* Just sanity check this.  Whine to user */
				else if (errno != EINTR)
					yell("Select failed with [%s]", 
						strerror(errno));
				break;
			}

			/* Nothing of note happened.  Punt */
			case 0:
				break;			/* Timeout */

			/* 
			 * Some file descriptor is ready.  
			 * Handle it and put data into buffers
			 */
			default:
			{
			    int	i;

			    for (i = 0; i <= global_max_fd; i++)
			    {
				if (FD_ISSET(i, &rd))
				{
				    if (x_debug & DEBUG_NEWIO)
				       yell("do_events: Handling a read event");
				    IO_READ(i) (i, 0, NULL);
				}
				else if (FD_ISSET(i, &wd))
				{
				    if (x_debug & DEBUG_NEWIO)
				      yell("do_events: Handling a write event");
				    IO_WRITE(i) (i, 0, NULL);
				}
			    }
			    break;
			}
		}
	}
	else if (x_debug & DEBUG_NEWIO)
		yell("do_events: Some data is ready.");

	if (x_debug & DEBUG_NEWIO)
		yell("do_events: Dispatching data");

	/* Actually dispatch the events now */
	logical__handle_data();

	if (x_debug & DEBUG_NEWIO)
		yell("do_events: do_events all done");
}

static void	event__track (int fd, Discipline *io_d)
{
	if (x_debug & DEBUG_NEWIO)
		yell("event__track [%d] called", fd);

	if (fd >= IO_ARRAYSIZE)
		panic("fd [%d] is bigger than IO_ARRAYSIZE [%d]",
			fd, IO_ARRAYSIZE);

	if (io_d->reader)
		FD_SET(fd, &readables);
	else
		FD_CLR(fd, &readables);

	if (io_d->writer)
		FD_SET(fd, &writeables);
	else
		FD_CLR(fd, &writeables);

	if (!IOX(fd))
		IOX(fd) = new_malloc(sizeof(Discipline));
	IO_READ(fd) 	= io_d->reader;
	IO_WRITE(fd) 	= io_d->writer;
	IO_EOF(fd)	= io_d->eof;
	IO_ERROR(fd)	= io_d->error;

	if (fd > global_max_fd)
		global_max_fd = fd;

	if (x_debug & DEBUG_NEWIO)
		yell("event__track [%d] ending", fd);
}

static void	event__untrack (int fd)
{
	FD_CLR(fd, &readables);
	FD_CLR(fd, &writeables);
	new_free((char **)&IOX(fd));
	close(fd);

	if (fd == global_max_fd)
	{
		while (!FD_ISSET(fd, &readables) && !FD_ISSET(fd, &writeables))
			fd--;
		global_max_fd = fd;
	}
}




/************************** STAGE 1B ************************************/
/************************* BUFFER DATA **********************************/
/*
 * This function returns 1 if an IO event can take place immediately.
 * This is true if there is dirty data in the read buffer or if the
 * write buffer is clean (meaning more data can be put into it).  Note
 * the contrary logic:  If there is no dirty data in the read buffer, or
 * if there IS dirty data in the write buffer, we need to do a select() 
 * call to wait until either data is ready to read(), or the data can be
 * write()n.
 */
static	int	buffer__data_waiting (void)
{
	int	i;

	for (i = 0; i < IO_ARRAYSIZE; i++)
	{
		if (READX(i) && RBUF_DIRTY(i) && IO_READ(i))
			return 1;
		if (WRITEX(i) && WBUF_DIRTY(i) == 0 && IO_WRITE(i))
			return 1;
	}
	return 0;
}



/*
 * Anyhow, this function read()s the data off the file descriptor and puts
 * it into the read buffer and marks the read buffer as dirty.  Presumably
 * then the client will call level2_handle_data which will pull data off of
 * the buffer and pass it along to the application.
 */
static	int	io__read (int fd, int unused, char *ignored)
{
	int	c;
	size_t	len;

	if (RBUF_DIRTY(fd))
		panic("fd [%d], buffer is dirty", fd);
	len = (RBUF_SIZE(fd) - (RBUF_END(fd) - RBUF_BUFFER(fd)) - 1);
	if (len <= 0)
		panic("fd [%d], buffer is full", fd);
		/* Probably shouldn't panic here. */

	c = read(fd, RBUF_END(fd), len);
	if (c == 0)
	{
		io_errno = -1;
		IO_EOF(fd) (fd);
		return 1;
	}
	else if (c < 0)
	{
		io_errno = errno;
		IO_ERROR(fd) (fd);
		return 1;
	}
	else
	{
		RBUF_DIRTY(fd) = 1;
		RBUF_END(fd) += c;
		return 0;
	}
}

/*
 * XXX The use of RBUF_DIRTY to pass new_fd is a hack, pure and simple.
 */
static int	io__accept (int fd, int unused, char *ignored)
{
	int	new_fd;
	int	sra = sizeof(ISA);

	if (RBUF_END(fd) != RBUF_BUFFER(fd))
		panic("Fd [%d], write != buffer", fd);

	new_fd = my_accept(fd, (SA *)&RBUF_BUFFER(fd), &sra);
	if (new_fd < 0)
	{
		IO_ERROR(fd) (fd);
		return 1;
	}
	else
	{
		RBUF_DIRTY(fd) = new_fd;
		RBUF_END(fd) += sra;
		return 0;
	}
}


/*
 * Anyhow, this function write()s the data in the buffer to the file descriptor
 * and if the buffer is empty, it marks the buffer as clean; otherwise the
 * buffer stays dirty.  Presumably the client will later call 
 * level2_handle_data and if the buffer is clean the application callback will
 * re-fill it for next time.
 */
static	int	io__write (int fd, int unused, char *ignored)
{
	int	c;
	size_t	len;

	if (!WBUF_DIRTY(fd))
		panic("fd [%d] is not dirty", fd);
	if (!(len = (size_t)(WBUF_END(fd) - WBUF_START(fd))))
		panic("Write buffer for fd (%d) is empty", fd);

	c = write(fd, WBUF_START(fd), len);
	if (c == 0)
	{
		io_errno = -1;
		IO_EOF(fd) (fd);
		return 1;
	}
	else if (c < 0)
	{
		io_errno = errno;
		IO_ERROR(fd) (fd);
		return 1;
	}
	else if (c < len)
	{
		WBUF_DIRTY(fd) = 1;
		WBUF_START(fd) += c;
		return 0;
	}
	else
	{
		WBUF_START(fd) = WBUF_END(fd) = WBUF_BUFFER(fd);
		WBUF_DIRTY(fd) = 0;
		return 0;
	}
}


static int	io__connect (int fd, int unused, char *ignored)
{
	WBUF_START(fd) = WBUF_END(fd) = WBUF_BUFFER(fd);
	WBUF_DIRTY(fd) = 0;
	return 0;
}



/****************************/
static	Discipline	io_read_write = {
	io__read,
	io__write,
	logical__eof,
	logical__error
};

static	Discipline	io_read_only = {
	io__read,
	NULL,
	logical__eof,
	logical__error
};

static	Discipline	io_write_only = {
	NULL,
	io__write,
	logical__eof,
	logical__error
};

static	Discipline	io_accept = {
	io__accept,
	NULL,
	NULL,
	logical__error
};

static	Discipline	io_connect = {
	NULL,
	io__connect,
	logical__eof,
	logical__error
};






/**************************** STAGE 2 **************************************/
/************************* LOGICAL HANDLERS ********************************/

static	int	logical__handle_data (void)
{
	int	i, called = 0;

	if (x_debug & DEBUG_NEWIO)
		yell("logical__handle_data called");

	for (i = 0; i < IO_ARRAYSIZE; i++)
	{
		if (!LOGICX(i))
			continue;

		if (LOGIC_READ(i))
		{
			if (RBUF_DIRTY(i))
			{
				called = 1;
				if (LOGIC_READ(i) (i, 0, NULL))
					continue;
			}
		}

		/* We need to check this -- it might have gone away! */
		if (!LOGICX(i))
			continue;

		if (LOGIC_WRITE(i))
		{
			if (WBUF_DIRTY(i) == 0)
			{
				called = 1;
				if (LOGIC_WRITE(i) (i, 0, NULL))
					continue;
			}
		}
	}

	if (x_debug & DEBUG_NEWIO)
		yell("logical__handle_data ending");
	return called;
}


/*
 * This is the basic function that registers a generic Level 3 Callback.
 * The 'logic' argument holds the Level 3 callbacks which are registered
 * with the Level 1 parser, and the 'appl' argument holds the Level 4
 * callbacks which are stashed in the Level e data about the 'fd file
 * descriptor.
 */
static	void	logical__track (int fd, Discipline *io, Discipline *logic, Discipline *appl)
{
	if (x_debug & DEBUG_NEWIO)
		yell("logical__track [%d] called", fd);

	if (fd >= IO_ARRAYSIZE)
		panic("fd [%d] is bigger than IO_ARRAYSIZE [%d]",
			fd, IO_ARRAYSIZE);

	if (!READX(fd))
		READX(fd) = new_malloc(sizeof(Buffer));
	if (!WRITEX(fd))
		WRITEX(fd) = new_malloc(sizeof(Buffer));
	if (!LOGICX(fd))
		LOGICX(fd) = new_malloc(sizeof(Discipline));
	if (!APPLX(fd))
		APPLX(fd) = new_malloc(sizeof(Discipline));

	RBUF_START(fd) = RBUF_END(fd) = RBUF_BUFFER(fd);
	RBUF_DIRTY(fd) = 0;
	WBUF_START(fd) = WBUF_END(fd) = WBUF_BUFFER(fd);
	WBUF_DIRTY(fd) = 0;

	LOGIC_READ(fd) = logic->reader;
	LOGIC_WRITE(fd) = logic->writer;
	LOGIC_EOF(fd) = logic->eof;
	LOGIC_ERROR(fd) = logic->error;

	APPL_READER(fd) = appl->reader;
	APPL_WRITER(fd) = appl->writer;
	APPL_EOF(fd) = appl->eof;
	APPL_ERROR(fd) = appl->error;

	event__track(fd, io);

	if (x_debug & DEBUG_NEWIO)
		yell("logical__track [%d] ending", fd);
}

/*
 * This removes all the Level 3 data about the 'fd' file descriptor and 
 * requests that the file descriptor no longer be queried for events.
 * Right now this is kind of handled by new_close, but that is a hack
 * and should be phased out.
 */
static	void	logical__untrack (int fd)	
{
	if (x_debug & DEBUG_NEWIO)
		yell("logical__untrack [%d] called", fd);

	if (!APPLX(fd) || !LOGICX(fd))
	{
		yell("Warning: logical__untrack called again on fd [%d]", fd);
		return;
	}

	new_free((char **)&READX(fd));
	new_free((char **)&WRITEX(fd));
	new_free((char **)&LOGICX(fd));
	new_free((char **)&APPLX(fd));
	event__untrack(fd);

	if (x_debug & DEBUG_NEWIO)
		yell("logical__untrack [%d] ending", fd);
}

int	new_close (int fd)
{
	if (fd < 0)
		return - 1;		/* Backwards compatability */
	logical__untrack(fd);
	return -1;
}

static int	logical__eof (int fd)
{
	if (APPL_EOF(fd))
		if (APPL_EOF(fd) (fd))
			return 1;
	logical__untrack(fd);
	return 0;
}

static int	logical__error (int fd)
{
	if (APPL_ERROR(fd))
		if (APPL_ERROR(fd) (fd))
			return 1;
	logical__untrack(fd);
	return 0;
}




/*********************** LEVEL 2 BUFFERING DISCIPLINES *********************/

/*************** THE DEFAULT DISCIPLINE **************/
/*
 * These "DEFAULT" handlers just act as a passthrough to the application's
 * discipline.  No data is changed, and all EOFs and errors are fatal unless
 * caught by the application callback.
 */
static int	logical__reader (int fd, int size, char *ignored)
{
	int	len = RBUF_END(fd) - RBUF_START(fd);
	char *	str;

	/* Defer as neccesary to application */
	if (APPL_READER(fd))
	{
		str = RBUF_START(fd);
		RBUF_DIRTY(fd) = 0;	/* XXX Hack? */
		RBUF_START(fd) = RBUF_END(fd) = RBUF_BUFFER(fd);
		if (APPL_READER(fd) (fd, len, str))
			return 1;
	}

	return 0;
}

static int	logical__writer (int fd, int size, char *unused)
{
	char	buffer[IOBUFSIZ + 1];
	int	c;

	/* Defer as neccesary to application */
	if (APPL_WRITER(fd))
	{
		if ((c = APPL_WRITER(fd) (fd, IOBUFSIZ, buffer)) > 0)
		{
			memcpy(WBUF_END(fd), buffer, c);
			WBUF_START(fd) = WBUF_BUFFER(fd);
			WBUF_END(fd) += c;
			WBUF_DIRTY(fd) = 1;
		}

		return 0;
	}

	/* Default action */
	else
		panic("logical__writer does not have a callback.");
}

static int	logical__accept (int fd, int size, char *ignored)
{
	int	retval = 0;
	int	fd2;
	char *	str;

	/* Defer as neccesary to application */
	if (APPL_READER(fd))
	{
		fd2 = RBUF_DIRTY(fd);
		str = RBUF_START(fd);
		RBUF_DIRTY(fd) = 0;
		RBUF_START(fd) = RBUF_END(fd) = RBUF_BUFFER(fd);

		retval = APPL_READER(fd) (fd, fd2, str);
	}
	else
		close(RBUF_DIRTY(fd));

	return retval;
}

/*************** THE READ/WRITE PASSTHROUGH DISCIPLINE *************/
/*
 * The NO BUFFERING I/O Discipline is a simple passthrough discipline.
 * All data is passed to and from the application layer as requested.
 * All default actions are taken as neccesary.
 *
 * Used by: DCC SEND, DCC GET, DCC RAW ($listen()), Screen Input.
 * Also used by nonblocking connects (who just set the writer function)
 */
static Discipline	no_buffer_logical = {
	logical__reader,
	logical__writer,
	NULL,
	NULL
};

int	no_buffering (int fd, Discipline *appl)
{
	logical__track(fd, &io_read_write, &no_buffer_logical, appl);
	return 0;
}

/******************** THE READ ONLY PASSTHROUGH DISCIPLINE **************/
/*
 * The NO BUFFERING READ ONLY I/O Discipline is a simple passthrough 
 * discipline that does not register a writeable callback.
 * All data is passed from the application layer as requested.
 * All default actions are taken as neccesary.
 *
 * Used by: DCC SEND, DCC GET, DCC RAW ($listen()), Screen Input.
 * Also used by nonblocking connects (who just set the writer function)
 */
static Discipline	no_buffer_ro_discipline = {
	logical__reader,
	NULL,
	NULL,
	NULL
};

int	no_buffering_readonly (int fd, Discipline *appl)
{
	logical__track(fd, &io_read_only, &no_buffer_ro_discipline, appl);
	return 0;
}



/************** THE EXEC PARTIAL BUFFERING DISCIPLINE *******************/
/*
 * The EXEC BUFFERED I/O Discipline will return data when one of the
 * following conditions is met:
 *	1) A newline is encountered
 *	2) The end of data is encountered.
 *
 * Used by exec.c
 */
static int	logical__exec_reader (int fd, int size, char *ignored)
{
	char	*line = NULL;
	char	*buffer;
	int	retval = 0;

	RBUF_END(fd)[0] = 0;
	buffer = RBUF_START(fd);
	if (!buffer || !*buffer)
		panic("No data to read in logical__exec_reader");

	if (!(line = next_line(buffer, &buffer)))
	{
		RBUF_DIRTY(fd) = 0;
		if (!APPL_READER(fd))
			panic("No logical__exec_reader callback");
		if (APPL_READER(fd) (fd, -strlen(buffer), buffer))
			return 1;
		return 0;
	}

	if (buffer < RBUF_END(fd))
		RBUF_DIRTY(fd) = 1;
	else
		RBUF_DIRTY(fd) = 0;

	if (!APPL_READER(fd))
		panic("No logical__exec_reader callback");
	retval = APPL_READER(fd) (fd, strlen(line), line);

	if (RBUF_DIRTY(fd) == 0)
		RBUF_END(fd) = RBUF_START(fd) = RBUF_BUFFER(fd);
	else
		RBUF_START(fd) = buffer;

	return 0;
}

static Discipline	exec_buffered_discipline = {
	logical__exec_reader,
	NULL,
	NULL,
	NULL
};

int	exec_buffering (int fd, Discipline *appl)
{
	logical__track(fd, &io_read_only, &exec_buffered_discipline, appl);
	return 0;
}


/************** THE FULLY LINE BUFFERED DISCIPLINE ********************/
/*
 * The LINE BUFFERED I/O Discipline will return data when one of the
 * following conditions is met:
 *	1) A newline is encountered
 *
 * Used by DCC CHAT and server.c
 */
static int	logical__line_reader (int fd, int size, char *ignored)
{
	char	*line;
	char	*buffer;

	RBUF_END(fd)[0] = 0;
	buffer = RBUF_START(fd);
	if (!buffer || !*buffer)
		panic("No data to read in logical__line_reader");

	/*
	 * If there is not a complete line in the buffer, copy what is
	 * there to the front of the buffer and punt.
	 */
	if (!(line = next_line(buffer, &buffer)))
	{
		RBUF_DIRTY(fd) = 0;
		ov_strcpy(RBUF_BUFFER(fd), RBUF_START(fd));
		RBUF_START(fd) = RBUF_BUFFER(fd);
		RBUF_END(fd) = RBUF_BUFFER(fd) + strlen(RBUF_BUFFER(fd));
		return 0;
	}

	/*
	 * Note that we can't change anything in the buffer here,
	 * because 'line' is still live and pointing to the next 
	 * line of data.  But what we can do is reset the neccesary
	 * pointers for next time.  If we have not yet drained the
	 * input buffer (the queue is dirty), then reset the START
	 * OF DATA pointer to the beginning of the next line.  The
	 * i/o stuff will immediately check to see if there is another
	 * line there, and if there is, it will be dispatched; otherwise
	 * if there is an imcomplete line, the above code will move it
	 * to the front of the queue where it can't do any damage.
	 * If the input buffer IS empty, however, then we want to reset
	 * the START and END pointers to the beginning of the buffer,
	 * so that the next read will start at the beginning of the
	 * buffer.  This fixes the bug where the input buffer could
	 * fill up, because we recieved multiple complete lines and
	 * START never got reset to the start of the queue.
	 */
	if (buffer < RBUF_END(fd))
	{
		RBUF_DIRTY(fd) = 1;
		RBUF_START(fd) = buffer;
	}
	else
	{
		RBUF_DIRTY(fd) = 0;
		RBUF_START(fd) = RBUF_END(fd) = RBUF_BUFFER(fd);
	}

	if (!APPL_READER(fd))
		panic("No logical__line_reader callback");
	if (APPL_READER(fd) (fd, strlen(line), line))
		return 1;

	return 0;
}

static Discipline	line_buffered_discipline = {
	logical__line_reader,
	NULL,
	NULL,
	NULL
};

int	line_buffering (int fd, Discipline *appl)
{
	if (x_debug & DEBUG_NEWIO)
		yell("line_buffering [%d] called", fd);

	logical__track(fd, &io_read_only, &line_buffered_discipline, appl);

	if (x_debug & DEBUG_NEWIO)
		yell("line_buffering [%d] ending", fd);
	return 0;
}

/************** THE NONBLOCKING CONNECT DISCIPLINE **********************/
/*
 * This utility function is used to stall until a nonblocking connect finishes
 */
static Discipline	nonblocking_discipline = {
	NULL,
	logical__writer,
	NULL,
	NULL
};

static Discipline	level3_nb_discipline = { NULL, NULL, NULL, NULL };

int	nonblocking_connect_callback (int fd, Writer writer)
{
	level3_nb_discipline.writer = writer;
	logical__track(fd, &io_connect, &nonblocking_discipline, 
				&level3_nb_discipline);
	WBUF_DIRTY(fd) = 1;
	return 0;
}

/************** THE WAITING FOR ACCEPT DISCIPLINE ************************/
/*
 * This discipline executes a callback when an accept() would succeed
 */
static Discipline	accept_discipline = {
	logical__accept,
	NULL,
	NULL,
	NULL
};

static Discipline	level3_a_discipline = { NULL, NULL, NULL, NULL };

int	accept_callback (int fd, Reader reader)
{
	level3_a_discipline.reader = reader;
	logical__track(fd, &io_accept, &accept_discipline, &level3_a_discipline);
	return 0;
}




/*************************** INIT STUFF ********************************/
void	set_up_newio (void)
{
	int	i;

#if 0
	x_debug |= DEBUG_NEWIO;
	x_debug |= DEBUG_INBOUND;
	x_debug |= DEBUG_OUTBOUND;
#endif

	IO_ARRAYSIZE = IO_ARRAYLEN;

	RESIZE(io__discipline, Discipline *, IO_ARRAYSIZE);
	RESIZE(logical__readbuf, Buffer *, IO_ARRAYSIZE);
	RESIZE(logical__writebuf, Buffer *, IO_ARRAYSIZE);
	RESIZE(logical__discipline, Discipline *, IO_ARRAYSIZE);
	RESIZE(application__discipline, Discipline *, IO_ARRAYSIZE);

	for (i = 0; i < IO_ARRAYSIZE; i++)
	{
		io__discipline[i] = NULL;
		logical__readbuf[i] = NULL;
		logical__writebuf[i] = NULL;
		logical__discipline[i] = NULL;
		application__discipline[i] = NULL;
	}
}

