/* $EPIC:$ */
/*
 * nutty.c: This file should prove once and for all that I HAVE gone 
 *	    off the deep end.
 *
 * Copyright  2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */
#include "irc.h"
#include "ircaux.h"
#include "vars.h"
#include "output.h"
#include "window.h"
#include "server.h"
#include "nutty.h"


/*
 * Here's the plan -- every line of output is described by a tuple of
 * information, containing:
 *
 *	(Window_refnum, from_server, who_from, who_level)
 *
 * Where "window_refnum" is the window that output is going to be sent to.
 * Where "server_refnum" is from_server, usually "window_refnum->server".
 * Where "who_from" is the person who sent a message, the channel that
 *		a message was sent to, or NULL otherwise.
 * Where "who_level" is the lastlog level that message is being sent out with.
 *		Typically "window_refnum->lastlog_level" includes it.
 *
 * It is useful to permit anyone to "intercept" any line of output, for
 * whatever purposes they have in mind (specifically, logging would do this.)
 * So we need to devise a way where any arbitrary function can be registered
 * to be called whenever any line is output.
 *
 * At the most basic of all levels, we have the fundamental callback system.
 * This system always invokes all of its callbacks for all lines of output,
 * passing the callback the four-tuple above, and the line of output.  What
 * that function does with that information is totaly up to it.  This is the
 * normal way to register a function if you want to see *ALL* output from the
 * *ENTIRE* client.
 *
 * In the intermediate level, we have a filter callback system.  This is a 
 * more practical level, since most callbacks are only interested in specific
 * types of output.  A callback function can register its interest in one
 * or more of the above types of the four-tuple (for example, just specify
 * the "who_from" as a channel, to log everything pertaining to the channel.)
 * and the intermediate level (which is itself registered with the fundamental
 * level) will keep track of its interest in ONLY when "who_from" is the same
 * as some specified channel, and invokes its callback each time.
 *
 * In the application level, we have the actual useful functions that do 
 * something for each type of output.  There are two fundamental callbacks:
 * the first is the infrastructure for the "filtering" callbacks, and the
 * second is just the global logfile handler.
 */

/*************************** LOW LEVEL STUFF ******************************/
/*
 * Each function may only be registered ONCE.  If you need multiple callbacks,
 * you need to create your own abstraction level.
 */
#define CSIZ	64
static	output_callback low_callbacks[CSIZ];	/* XXX Limited for now XXX */
static	int	callback_size = -1;

/*
 * Anyone may call this function at any time to register a function to be
 * called back for every line of output.  You may only register a function
 * ONCE.  If you attempt to re-register a function, -1 will be returned.
 * If there is no more room to register new functions, -1 is returned.
 * The artificial limitation on the number of functions that may be registered
 * will be removed some time soon.  The function returns 0 on success.
 */
int	register_output_callback (output_callback func)
{
	int	i;

	if (callback_size == -1)
	{
		for (i = 0; i < CSIZ; i++)
			low_callbacks[i] = NULL;
		callback_size = 0;
	}

	for (i = 0; i < CSIZ; i++)
	{
		if (low_callbacks[i] && low_callbacks[i] == func)
			return -1;		/* Already registered */
	}
	for (i = 0; i < CSIZ; i++)
	{
		if (low_callbacks[i] == NULL)
		{
			low_callbacks[i] = func;
			return 0;		/* Registration ok */
		}
	}
	return -1;				/* No more room */
}

int	unregister_output_callback (output_callback func)
{
	int	i;

	for (i = 0; i < CSIZ; i++)
	{
		if (low_callbacks[i] && low_callbacks[i] == func)
		{
			low_callbacks[i] = NULL;
			return 0;	/* Now unregistered */
		}
	}
	return -1;			/* Not registered */
}

void	do_output_callbacks (Winref w, Servref s, Char *wf, int wl, const u_char *output)
{
	output_info	info;
	int		i;

	info.winref = w;
	info.servref = s;
	info.who_from = wf;
	info.who_level = wl;

	for (i = 0; i < 64; i++)
	{
		if (low_callbacks[i])
			low_callbacks[i](&info, output);
	}
}

#if 0
/**************************** INTERMEDIATE STUFF ***************************/
static void	filter_callback (const output_info *info, const u_char *output)
{
	int	i;

	for (i = 0; i < num_logs; i++)
	{
		output_info	*i;

		i = &med_callbacks[i].info;
		if (i->winref != -1 && i->winref != info->winref)
			continue;
		if (i->servref != -1 && i->servref != info->servref)
			continue;
		if (i->who_from && !my_stricmp(i->who_from, info->who_from))
			continue;
		if (i->who_level && (i->who_level & info->who_level == 0))
			continue;

		med_callbacks[i].func(&info, output);
	}
}

int	call_when_target_output (output_callback func, Char *target)
{
	output_info	info;

	info.winref = -1;
	info.servref = -1;
	info.who_from = target;
	info.who_level = 0;
	return call_when_this_output(func, &info);
}

int	call_when_level_output (output_callback func, int level)
{
	output_info	info;

	info.winref = -1;
	info.servref = -1;
	info.who_from = NULL;
	info.who_level = level;
	return call_when_this_output(func, &info);
}

int	call_when_window_output (output_callback func, Winref w)
{
	output_info	info;

	info.winref = w;
	info.servref = -1;
	info.who_from = NULL;
	info.who_level = 0;
	return call_when_this_output(func, &info);
}

int	call_when_server_output (output_callback func, Servref s)
{
	output_info	info;

	info.winref = -1;
	info.servref = s; 
	info.who_from = NULL;
	info.who_level = 0;
	return call_when_this_output(func, &info);
}

int	call_when_any_output (output_callback func, output_info *ignored)
{
	output_info	info;

	info.winref = -1;
	info.servref = -1; 
	info.who_from = NULL;
	info.who_level = 0;
	return call_when_this_output(func, &info);
}

int	call_when_this_output (output_callback func, output_info *info)
{
	output_info	info;
}

void	cancel_callback (int callnum)
{
}

#endif
