/* $EPIC:$ */
/*
 * output.c: window independant output routines.
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#define __need_term_flush__
#define __need_putchar_x__
#include "irc.h"
#include "output.h"
#include "vars.h"
#include "input.h"
#include "term.h"
#include "lastlog.h"
#include "window.h"
#include "screen.h"
#include "hook.h"
#include "ctcp.h"
#include "ircaux.h"
#include "alias.h"
#include "server.h"
#include "list.h"
#include "parse.h"
#include "names.h"
#include "nutty.h"
#include "log.h"
#include <sys/stat.h>

/*
 * When some code wants to override the default lastlog level, and needs
 * to have some output go into some explicit window (such as for /xecho -w),
 * then while to_window is set to some window, *ALL* output goes to that
 * window.  Dont forget to reset it to NULL when youre done!  ;-)
 */
	static	Window	*to_window = NULL;

/*
 *
 */
	static	Window	*target_window = NULL;

/*
 * This is used to note who the currently processed message was from.
 * Since each window has (through /window add) the ability to "grab" all
 * of the output from a given nickname, we store the nickname here and then
 * refer to it when anything is outputted.
 */
	static const char *who_from = (char *) 0;	

/*
 * This is the lastlog level that any output should be sent out at.  This
 * determines what window output ultimately ends up in.
 */
	static	int	who_level = LOG_CRAP;


/* make this buffer *much* bigger than needed */
#define OBNOXIOUS_BUFFER_SIZE BIG_BUFFER_SIZE * 10
static	char	putbuf[OBNOXIOUS_BUFFER_SIZE + 1];

/* 
 * unflash: sends a ^[c to the screen
 * Must be defined to be useful, cause some vt100s really *do* reset when
 * sent this command. >;-)
 * Now that you can send ansi sequences, this is much less inportant.. 
 */
void	unflash (void)
{
#ifdef HARD_UNFLASH
	tputs_x("\033c");		/* hard reset */
#else
	tputs_x("\033)0");		/* soft reset */
#endif
}

/* sig_refresh_screen: the signal-callable versino of refresh_screen */
SIGNAL_HANDLER(sig_refresh_screen)
{
	refresh_a_screen(main_screen);
}

/*
 * refresh_screen: Whenever the REFRESH_SCREEN function is activated, this
 * swoops into effect 
 */
void 	refresh_screen (char dumb, char *dumber)
{
	refresh_a_screen(current_window->screen);
}

void	refresh_a_screen (Screen *screen)
{
	output_screen = screen;

	unflash();
	term_clear_screen();

	if (screen != main_screen || term_resize())
		recalculate_windows(screen);
	else
		redraw_all_windows();

	if (need_redraw)
		need_redraw = 0;

	update_all_windows();
	update_input(UPDATE_ALL);

	output_screen = NULL;
}


/*
 * extern_write -- set access permissions for our terminal
 * This implemenation was taken from 4.4BSD.
 */
BUILT_IN_COMMAND(extern_write)
{
	char *tty;
	struct stat sbuf;
	const int OTHER_WRITE = 020;

	if (!(tty = ttyname(2)))
	{
		yell("Could not figure out the name of your tty device!");
		return;
	}
	if (stat(tty, &sbuf) < 0)
	{
		yell("Could not get the information about your tty device!");
		return;
	}
	if (!args || !*args)
	{
		if (sbuf.st_mode & 020)
			say("Mesg is 'y'");
		else
			say("Mesg is 'n'");
		return;
	}
	switch (args[0])
	{
		case 'y' :
		{
			if (chmod(tty, sbuf.st_mode | OTHER_WRITE) < 0)
			{
				yell("Could not set your tty's mode!");
				return;
			}
			break;
		}
		case 'n' :
		{
			if (chmod(tty, sbuf.st_mode &~ OTHER_WRITE) < 0)
			{
				yell("Could not set your tty's mode!");
				return;
			}
			break;
		}
		default :
			say("Usage: /%s [Y|N]", command);
	}
}

/* Bootstrap the terminal, screen, and window system */
int	init_display (void)
{
	/*
	 * System dependant stuff
	 */
	if (term_init())
		return -1;

	term_clear_screen();
	term_resize();

	/*
	 * System independant stuff
	 */
	create_new_screen();
	new_window(main_screen);
	update_all_windows();

	/*
	 * Put the cursor on the screen.
	 */
	term_move_cursor(0, 0);
	return 0;
}

/*
 * put_it: the primary irc display routine.  This routine can be used to
 * display output to a user window.  Its ok to have newlines or tabs in
 * the output, but you should be careful not to overflow the 10k buffer
 * used to hold the output (use display if you just want to output an
 * unbounded string).
 *
 * Some systems (notorously Ultrix) cannot gracefully convert float
 * variables through "..." into va_list, and attempting to do so will 
 * cause a crash on dereference (in vsprintf()), so do do that.  You'll
 * have to sprintf() the floats into strings and then pass the string 
 * in with %s.  Ugh.  This is not a bug on our part.
 */
void	vput_it (const char *format, va_list args)
{
	if (window_display == 1 && format)
	{
		vsnprintf(putbuf, OBNOXIOUS_BUFFER_SIZE, format, args);
		display(putbuf);
	}
}


void 	put_it (const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vput_it(format, args);
	va_end(args);
}

/* 
 * This is an alternative form of put_it which writes three asterisks
 * before actually putting things out.
 */
void 	vsay (const char *format, va_list args)
{
	if (window_display == 1 && format)
	{
		char *str;

		*putbuf = 0;
		if ((str = get_string_var(BANNER_VAR)))
		{
			if (get_int_var(BANNER_EXPAND_VAR))
			{
			    int af;
			    char *foo;

			    foo = expand_alias(str, empty_string, &af, NULL);
			    strlcpy(putbuf, foo, OBNOXIOUS_BUFFER_SIZE);
			    new_free(&foo);
			}
			else
			    strlcpy(putbuf, str, OBNOXIOUS_BUFFER_SIZE);

			strlcat(putbuf, " ", OBNOXIOUS_BUFFER_SIZE);
		}

		vsnprintf(putbuf + strlen(putbuf), 
			OBNOXIOUS_BUFFER_SIZE - strlen(putbuf) - 1, 
			format, args);

		display(putbuf);
	}
}

void	say (const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vsay(format, args);
	va_end(args);
}

void	yell (const char *format, ...)
{
	if (format)
	{
		va_list args;
		va_start (args, format);
		vsnprintf(putbuf, OBNOXIOUS_BUFFER_SIZE, format, args);
		va_end(args);
		if (do_hook(YELL_LIST, "%s", putbuf))
			display(putbuf);
	}
}

/*
 * Error is exactly like yell, except that if the error occured while
 * you were loading a script, it tells you where it happened.
 */
void 	error (const char *format, ...)
{
	dump_load_stack(0);
	if (format)
	{
		va_list args;
		va_start (args, format);
		vsnprintf(putbuf, OBNOXIOUS_BUFFER_SIZE, format, args);
		va_end(args);
		do_hook(YELL_LIST, "%s", putbuf);
		display(putbuf);
	}
}

/*
 * EVERY LAST BIT OF OUTPUT IN THE ENTIRE CLIENT siphons through this
 * function.  Thus, any universal rules that must be used must be put
 * into this function.  "display" and "display_to" are merely wrapper
 * functions that determine which window they want to output to.  This
 * is the function that causes everything to happen from there.
 */
static 	void	universal_display (Window *w, const unsigned char *buffer)
{
	do_output_callbacks(w->refnum, from_server, who_from, 
				who_level, buffer);
	w->add(w, buffer);
}

/*
 * display: This adds the given null terminated buffer to the screen.
 * That is, it routes the line to the appropriate window.  It also handles
 * /redirect handling.
 */
void 	display (const unsigned char *buffer)
{
	Window *w;

	if (in_window_command)
	{
		in_window_command = 0;	/* Inhibit looping! */
		update_all_windows();
		in_window_command = 1;
	}

	if (!current_window)
	{
		puts(buffer);
		return;
	}

	if (dumb_mode)
	{
		add_to_lastlog(current_window, buffer);
		if (do_hook(WINDOW_LIST, "%u %s", 
				current_window->refnum, buffer))
			puts(buffer);
		fflush(stdout);
		return;
	}

	if (to_window)
		w = to_window;
	else if (target_window)
		w = target_window;
	else
		w = current_window;

	universal_display(w, buffer);
}

void	display_to (Winref win, const unsigned char *buffer)
{
	Window *w;

	if (!(w = get_window(win)))
		return;			/* Whatever */

	universal_display(w, buffer);
}

void 	display_raw (const unsigned char *buffer)
{
	Screen 	*old;

	if (!current_window)
	{
		puts(buffer);
		return;
	}

	old = output_screen;
	if (dumb_mode)
		output_screen = current_window->screen;
	else if (to_window)
		output_screen = to_window->screen;
	else if (target_window)
		output_screen = target_window->screen;
	else
		output_screen = current_window->screen;

	tputs_x(buffer);
	term_flush();
	output_screen = old;

	return;
}

/***************************************************************************/
/*
 * message_to: This allows you to specify a window (by refnum) as a
 * destination for messages.  Used by EXEC routines quite nicely 
 */
Winref 	message_to (Winref refnum)
{
	Winref	old = -1;

	if (to_window)
		old = to_window->refnum;
	to_window = (refnum >= 0) ? get_window(refnum) : NULL;
	return old;
}

/*
 * save_message_from: this is used to save (for later restoration) the
 * who_from variable.  This is needed when a function (do_hook) is about 
 * to call another function (parse_line) it knows will change who_from.
 * The values are saved on the stack so it will be recursive-safe.
 *
 * NO CHEATING when you call this function to get the value of who_from! ;-)
 */
void 	save_display_target (const char **saved_from, int *saved_level)
{
	*saved_from = who_from;
	*saved_level = who_level;
}

/* restore_message_from: restores a previously saved who_from variable */
void 	restore_display_target (const char *saved_from, int saved_level)
{
	who_from = saved_from;
	who_level = saved_level;
}

void 	set_display_target (const char *who, int level)
{
	Window	*tmp;
	Winref	w;

#ifdef NO_CHEATING
	malloc_strcpy(&who_from, who);
#else
	who_from = who;
#endif
	who_level = level;


	/*
	 * Now we try to find the window that this output level would
	 * be directed to.  This was transplanted from the old add_to_screen().
	 */

	/*
	 * LOG_CURRENT means everything goes to the current window.
	 */
	if (who_level == LOG_CURRENT && current_window->server == from_server)
	{
		target_window = current_window;
		return;
	}

	/*
	 * Next priority is to honor who_from (using /window bind, 
	 * /window channel, or /window add)
	 */
	if (who_from)
	{
		if (is_channel(who_from) && from_server != -1)
		{
			if ((w = channel_window(who_from, from_server)))
			{
				tmp = get_window(w);
				target_window = tmp;
				return;
			}
		}
	
		tmp = NULL;
		while (traverse_all_windows(&tmp))
		{
#if 0		/* This should be completely unnecesary now */
			/*
			 * Check for /WINDOW CHANNELs that apply.
			 * (Any current channel will do)
			 */
			if (tmp->server == from_server)
			{
				Char	*cc;

				cc = get_window_channel(tmp->refnum);
				if (cc && !my_stricmp(who_from, cc))
				{
					target_window = tmp;
					return;
				}
			}
			else
#endif

			/*
			 * Check for /WINDOW QUERYs that apply.
			 */
			    if (tmp->query_nick &&
			   ( ((who_level == LOG_MSG || who_level == LOG_NOTICE
			    || who_level == LOG_DCC || who_level == LOG_CTCP
			    || who_level == LOG_ACTION)
				&& !my_stricmp(who_from, tmp->query_nick)
				&& from_server == tmp->server)
			  || ((who_level == LOG_DCC || who_level == LOG_CTCP
			    || who_level == LOG_ACTION)
				&& *tmp->query_nick == '='
				&& !my_stricmp(who_from, tmp->query_nick + 1))
			  || ((who_level == LOG_DCC || who_level == LOG_CTCP
			    || who_level == LOG_ACTION)
				&& *tmp->query_nick == '='
				&& !my_stricmp(who_from, tmp->query_nick))))
			{
				target_window = tmp;
				return;
			}
		}

		/*
		 * Check for /WINDOW NICKs that apply
		 */
		tmp = NULL;
		while (traverse_all_windows(&tmp))
		{
			if (tmp->nicks && from_server == tmp->server)
			{
				if (find_in_list((List **)&(tmp->nicks), 
					who_from, !USE_WILDCARDS))
				{
					target_window = tmp;
					return;
				}
			}
		}
	}

	/*
	 * Check to see if this level should go to current window
	 */
	if ((current_window_level & who_level) && 
		(who_level == LOG_DCC || 
		(current_window->server == from_server)))
	{
		target_window = current_window;
		return;
	}

	/*
	 * Check to see if any window can claim this level
	 */
	tmp = NULL;
	while (traverse_all_windows(&tmp))
	{
		/*
		 * Check for /WINDOW LEVELs that apply
		 */
		if (((from_server == tmp->server) || (from_server == -1)) &&
		    (who_level & tmp->window_level))
		{
			target_window = tmp;
			return;
		}
	}

	/*
	 * If all else fails, if the current window is connected to the
	 * given server, use the current window.
	 */
	if (from_server == current_window->server)
	{
		target_window = current_window;
		return;
	}

	/*
	 * And if that fails, look for ANY window that is bound to the
	 * given server (this never fails if we're connected.)
	 */
	tmp = NULL;
	while (traverse_all_windows(&tmp))
	{
		if (tmp->server == from_server)
		{
			target_window = tmp;
			return;
		}
	}

	/*
	 * No window found for a server is usually because we're
	 * disconnected or not yet connected.
	 */
	target_window = current_window;
	return;
}

void	reset_display_target (void)
{
	set_display_target(NULL, LOG_CRAP);
}

int	output_level (void)
{
	return who_level;
}

const char *	output_target (void)
{
	return who_from;
}

