/* $EPIC:$ */
/*
 * parse.c: handles messages from the server.   Believe it or not.  I
 * certainly wouldn't if I were you. 
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "server.h"
#include "names.h"
#include "vars.h"
#include "ctcp.h"
#include "hook.h"
#include "commands.h"
#include "ignore.h"
#include "lastlog.h"
#include "ircaux.h"
#include "crypt.h"
#include "term.h"
#include "flood.h"
#include "window.h"
#include "screen.h"
#include "output.h"
#include "numbers.h"
#include "parse.h"
#include "notify.h"
#include "notice.h"
#include "timer.h"

#define STRING_CHANNEL 	'+'
#define MULTI_CHANNEL 	'#'
#define LOCAL_CHANNEL 	'&'
#define ID_CHANNEL	'!'

#define space 		' '	/* Taken from rfc 1459 */
#define	MAXPARA		15	/* Taken from rfc 1459 */

static  void    strip_modes (char *, char *, char *);

/* User and host information from server 2.7 */
const 	char	*FromUserHost = NULL;

/* fake: alerts the user. */
void 	fake (void)
{
	yell("-- Fake Message receieved -- ");
	return;
}

/*
 * is_channel: determines if the argument is a channel.  If it's a number,
 * begins with MULTI_CHANNEL and has no '*', or STRING_CHANNEL, then its a
 * channel 
 */
int 	is_channel(const char *to)
{
	return ( (to) && (     (*to == MULTI_CHANNEL) 
			    || (*to == STRING_CHANNEL)
		            || (*to == LOCAL_CHANNEL) 
			    || (*to == ID_CHANNEL)));
}


char *	PasteArgs (char **arg_list, int paste_point)
{
	int	i;

	/*
	 * Make sure there are enough args to parse...
	 */
	for (i = 0; i < paste_point; i++)
		if (!arg_list[i])
			return NULL;		/* Not enough args */

	for (i = paste_point; arg_list[i] && arg_list[i + 1]; i++)
		arg_list[i][strlen(arg_list[i])] = ' ';
	arg_list[paste_point + 1] = NULL;
	return arg_list[paste_point];
}

/*
 * BreakArgs: breaks up the line from the server, in to where its from,
 * setting FromUserHost if it should be, and returns all the arguements
 * that are there.   Re-written by phone, dec 1992.
 *		     Re-written again by esl, april 1996.
 *
 * This doesnt strip out extraneous spaces any more.
 */
void 	BreakArgs (char *Input, char **Sender, char **OutPut)
{
	int	ArgCount;
	char *	fuh;

	/*
	 * Paranoia.  Clean out any bogus ptrs still left on OutPut...
	 */
	for (ArgCount = 0; ArgCount < MAXPARA; ArgCount++)
		OutPut[ArgCount] = NULL;
	ArgCount = 0;

	/*
	 * The RFC describes it fully, but in a short form, a line looks like:
	 * [:sender[!user@host]] COMMAND ARGUMENT [[:]ARGUMENT]{0..14}
	 */

	/*
	 * Look to see if the optional :sender is present.
	 */
	if (*Input == ':')
	{
		*Sender = ++Input;
		while (*Input && *Input != space)
			Input++;
		if (*Input == space)
			*Input++ = 0;

		/*
		 * Look to see if the optional !user@host is present.
		 */
		fuh = *Sender;
		while (*fuh && *fuh != '!')
			fuh++;
		if (*fuh == '!')
			*fuh++ = 0;
	}
	/*
	 * No sender present.
	 */
	else
		*Sender = fuh = empty_string;

	FromUserHost = fuh;

	/*
	 * Now we go through the argument list...
	 */
	for (;;)
	{
		while (*Input && *Input == space)
			Input++;

		if (!*Input)
			break;

		if (*Input == ':')
		{
			OutPut[ArgCount++] = ++Input;
			break;
		}

		OutPut[ArgCount++] = Input;
		if (ArgCount >= MAXPARA)
			break;

		while (*Input && *Input != space)
			Input++;
		if (*Input == space)
			*Input++ = 0;
	}
	OutPut[ArgCount] = NULL;
}

/* in response to a TOPIC message from the server */
static void 	p_topic (char *from, char **ArgList)
{
	char	*high;

	if (!ArgList[1]) 
	{
		fake();
		return;
	}

	switch (check_ignore_channel(from, FromUserHost, 
					ArgList[0], IGNORE_TOPICS))
	{
		case IGNORED:
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
			break;
	}

	if (!new_check_flooding(from, ArgList[0], ArgList[1], TOPIC_FLOOD))
		return;

	set_display_target(ArgList[0], LOG_CRAP);
	if (do_hook(TOPIC_LIST, "%s %s %s", from, ArgList[0], ArgList[1]))
		say("%s has changed the topic on channel %s to %s",
			from, channel_friendly(ArgList[0], -1), ArgList[1]);
	reset_display_target();
}

static void 	p_wallops (char *from, char **ArgList)
{
	int 	server_wallop = strchr(from, '.') ? 1 : 0;

	if (!ArgList[0])
	{
		fake();
		return;
	}

	/* wallops from a server */
	if (server_wallop || check_flooding(from, WALLOP_FLOOD, ArgList[0]))
	{
		char	*high;

		switch (check_ignore(from, FromUserHost, IGNORE_WALLOPS))
		{
			case (IGNORED):
				return;
			case (HIGHLIGHTED):
				high = highlight_char;
				break;
			default:
				high = empty_string;
				break;
		}

		set_display_target(from, LOG_WALLOP);
		if (do_hook(WALLOP_LIST, "%s %c %s", 
					from, 
					server_wallop ? 'S' : '*', 
					ArgList[0]))
			put_it("%s!%s%s!%s %s", 
					high, from, 
					server_wallop ? empty_string : star, 
					high, ArgList[0]);

		if (beep_on_level & LOG_WALLOP)
			beep_em(1);
		reset_display_target();
	}
}

static void 	p_privmsg (char *from, char **Args)
{
	int		list_type,
			flood_type,
			log_type;
	char		*flood_channel = NULL;
	unsigned char	ignore_type;
	char		*ptr,
			*to;
	char		*high;
	char		*target;

	PasteArgs(Args, 1);
	to = Args[0];
	ptr = Args[1];
	if (!to || !ptr)
	{
		fake();
		return;
	}

	/*
	 * Do ctcp's first, and if there's nothing left, then dont
	 * go to all the work below.  Plus, we dont set the display target
	 * until we know there's other stuff besides the ctcp in the
	 * message, which keeps things going to the wrong window.
	 */
	ptr = do_ctcp(from, to, ptr);
	if (!*ptr)
		return;

	/* ooops. cant just do is_channel(to) because of # walls... */
	if (is_channel(to) && im_on_channel(to, from_server))
	{
		target = to;

		set_server_public_nick(from);
		flood_channel = to;
		flood_type = PUBLIC_FLOOD;
		log_type = LOG_PUBLIC;
		ignore_type = IGNORE_PUBLIC;

		/* channel_nomsgs is an optimization hack to save a lookup */
		if (channel_nomsgs(to, from_server) == 0 && 
		    channel_nick_onchannel(to, from_server, from) == 0)
			list_type = PUBLIC_MSG_LIST;
		else
		{
			if (is_current_channel(to, from_server))
				list_type = PUBLIC_LIST;
			else
				list_type = PUBLIC_OTHER_LIST;
		}
	}
	else
	{
		target = from;

		if (!is_me(-1, to))
		{
			log_type = LOG_WALL;
			ignore_type = IGNORE_WALLS;
			list_type = MSG_GROUP_LIST;
			flood_type = WALL_FLOOD;
		}
		else
		{
			log_type = LOG_MSG;
			ignore_type = IGNORE_MSGS;
			list_type = MSG_LIST;
			flood_type = MSG_FLOOD;
		}
	}

	set_doing_privmsg(1);

	switch (check_ignore_channel(from, FromUserHost, to, ignore_type))
	{
		case IGNORED:
			set_doing_privmsg(0);
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
			break;
	}


	if (!new_check_flooding(from, flood_channel, ptr, flood_type))
	{
		set_doing_privmsg(0);
		return;
	}

	if (sed && !do_hook(ENCRYPTED_PRIVMSG_LIST,"%s %s %s", from, to, ptr))
	{
		sed = 0; 
		set_doing_privmsg(0);
		return;
	}

	set_display_target(target, LOG_MSG);
	switch (list_type)
	{
	    case PUBLIC_MSG_LIST:
	    {
		if (do_hook(list_type, "%s %s %s", from, to, ptr))
		    put_it("%s(%s/%s)%s %s", 
			high, from, channel_friendly(to, -1), high, ptr);
		break;
	    }

	    case MSG_GROUP_LIST:
	    {
		if (do_hook(list_type, "%s %s %s", from, to, ptr))
		    put_it("%s<-%s:%s->%s %s", 
			high, from, channel_friendly(to, -1), high, ptr);
		break;
	    }

	    case MSG_LIST:
	    {
		set_server_recv_nick(from);
		if (get_server_away(-2))
			beep_em(get_int_var(BEEP_WHEN_AWAY_VAR));

		if (do_hook(list_type, "%s %s", from, ptr))
		{
			if (get_server_away(-2))
			{
				put_it("%s*%s*%s %s <%.16s>", 
					high, from, high, ptr, 
					my_ctime(time(NULL)));
			}
			else
				put_it("%s*%s*%s %s", high, from, high, ptr);
		}
		break;
	    }

	    case PUBLIC_LIST:
	    {
		if (do_hook(list_type, "%s %s %s", from, to, ptr))
			put_it("%s<%s>%s %s", high, from, high, ptr);
		break;
	    }

	    case PUBLIC_OTHER_LIST:
	    {
		if (do_hook(list_type, "%s %s %s", from, to, ptr))
			put_it("%s<%s:%s>%s %s", 
				high, from, channel_friendly(to, -1), high, ptr);
		break;
	    }
	}
	reset_display_target();

	sed = 0;
	set_doing_privmsg(0);
	if (beep_on_level & log_type)
		beep_em(1);
	return;
}

static void 	p_quit (char *from, char **ArgList)
{
	int		one_prints = 0;
	int		i, num;
	char *		Reason;
	Channel_List	chanlist;

	notify_mark(from, 0, 0);

	if (check_ignore(from, FromUserHost, IGNORE_CRAP) == IGNORED)
		return;

	PasteArgs(ArgList, 0);
	Reason = ArgList[0] ? ArgList[0] : "?";

	num = channel_getlist(from_server, from, chanlist);

	/* Probably should do this AFTER we get the channels, eh? */
	channel_nick_remove(NULL, from_server, from);

	if (num)
	{
		for (i = 0; i < num; i++)
		{
			set_display_target(chanlist[i], LOG_CRAP);
			if (do_hook(CHANNEL_SIGNOFF_LIST, "%s %s %s",
					chanlist[i], from, Reason))
				one_prints = 1;
		}
	}

	if (one_prints)
	{
		if (num)
			set_display_target(chanlist[0], LOG_CRAP);
		else
			set_display_target(NULL, LOG_CRAP);

		if (do_hook(SIGNOFF_LIST, "%s %s", from, Reason))
			say("Signoff: %s (%s)", from, Reason);
	}

	reset_display_target();
}

static void 	p_pong (char *from, char **ArgList)
{
	PasteArgs(ArgList, 1);

	if (check_ignore(from, FromUserHost, IGNORE_CRAP) == IGNORED)
		return;

	if (do_hook(PONG_LIST, "%s %s %s", from, ArgList[0], ArgList[1])
		&& ArgList[0] && strchr(ArgList[0], '.'))
	    say("%s: PONG received from %s (%s)", 
			ArgList[0], from, ArgList[1]);
}

static void 	p_error (char *from, char **ArgList)
{
	PasteArgs(ArgList, 0);
	if (!ArgList[0])
		{fake();return;}
	say("%s", ArgList[0]);
}


static void 	p_channel (char *from, char **ArgList)
{
	char	*c, 
		*channel = NULL;
	int 	op = 0, 
		vo = 0, 
		ha = 0;
	char 	extra[20];
	char	*high;

	if (!strcmp(ArgList[0], zero))
		{fake();return;}

	channel = ArgList[0];
	set_server_joined_nick(from);

	/* Handle the bogus protocol change in ircnet2.9 */
	if ((c = strchr(channel, '\007')))
	{
		for (*c++ = 0; *c; c++)
		{
			     if (*c == 'o') op = 1;
			else if (*c == 'v') vo = 1;
		}
	}

	if (is_me(from_server, from))
		channel_status_change(channel, from_server, SYNCING);
	else
	{
		channel_nick_add(channel, from_server, from, 0, op, vo, ha);
		channel_userhost_add(channel, from_server, from, FromUserHost);
	}

	/* Check for ignore */
	switch (check_ignore_channel(from, FromUserHost, channel, IGNORE_JOINS))
	{
		case IGNORED:
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
			break;
	}

	/* Check for floods */
	if (!new_check_flooding(from, channel, star, JOIN_FLOOD))
		return;

	*extra = 0;
	if (op)
		strcat(extra, " (+o)");
	if (vo)
		strcat(extra, " (+v)");

	set_display_target(channel, LOG_CRAP);
	if (do_hook(JOIN_LIST, "%s %s %s %s", 
			from, channel, FromUserHost, extra))
		say("%s%s%s (%s) has joined channel %s%s%s%s", 
			high, from, high, FromUserHost, 
			high, channel_friendly(channel, -1), high, extra);
	reset_display_target();

	/*
	 * This should be done last to ensure that the userhost has been
	 * properly handled...
	 */
	notify_mark(from, 1, 0);
}


static void 	p_invite (char *from, char **ArgList)
{
	char	*high;

	if (!ArgList[0] || !ArgList[1])
		return;

	set_server_invite_channel(ArgList[1]);
	set_server_recv_nick(from);

	switch (check_ignore_channel(from, FromUserHost, 
					ArgList[1], IGNORE_INVITES))
	{
		case IGNORED:
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
			break;
	}

	if (!check_flooding(from, INVITE_FLOOD, ArgList[1]))
		return;

	set_display_target(from, LOG_CRAP);
	if (do_hook(INVITE_LIST, "%s %s %s", 
			from, ArgList[1],FromUserHost))
		say("%s%s (%s)%s invites you to channel %s", 
			high, from, FromUserHost, high, ArgList[1]);
	reset_display_target();
}

/* 
 * Reconnect has been put back in the "jon2 comprimise" 
 *
 * Unlimited autoreconnect for server kills
 * Limited autoreconnect for forground procis on oper kills
 *	(/set auto_reconnect ON)
 * No autoreconnect on oper kills for background procis
 *
 * /on disconnect is still always hooked no matter what.
 *
 * You may configure this any way you please.. I really dont give
 * a darn what autoconnect you use... but this is a fair default.
 *
 * NOTE: If you want to change it, you *will* have to do it yourself.
 * Im quite serious.  Im not going to change this.
 */
static void 	p_kill (char *from, char **ArgList)
{
	char *	victim;
	char *	bastard;
	char *	reason;

	/*
	 * Yes. these copies really _are_ neccesary and you should not ever
	 * take them out.  Both 'from' and 'ArgList' point into the io buffer
	 * that is maintained by newio2.c on behalf of the server object, and
	 * when we call close_server() down below, that buffer is revoked,
	 * and all the pointers into it are stale.  We must never _EVER_ touch
	 * either `from' or 'ArgList' after a call to close_server().  Ever.
	 */
	bastard = LOCAL_COPY(from);
	victim = LOCAL_COPY(ArgList[0]);
	if (!(reason = ArgList[1]))
		reason = "No reason given";
	reason = LOCAL_COPY(reason);

	/*
	 * Bogorific Microsoft Exchange ``IRC'' server sends out a KILL
	 * protocol message instead of a QUIT protocol message when
	 * someone is killed on your channel.  Do the obviously appropriate
	 * thing and reroute this misdirected protocol message to p_quit,
	 * where it should have been sent in the first place.  
	 * Die Microsoft, Die.
	 */
	if (!is_me(-1, victim))
	{
		/* I don't care if this doesn't work. */
		p_quit(from, ArgList);		/* Die Microsoft, Die */
		return;
	}

	/*
	 * We've been killed.  Close up shop on this server. 
	 */
	close_server(from_server, empty_string);

	/*
	 * Server Kill? (eg, Nick Collision?)
	 */
	if (strchr(bastard, '.'))
        {
		say("Server [%s] has rejected you (probably due to a "
			"nick collision)", bastard);
		get_connected(from_server, from_server);
		return;
	}


	/*
	 * Killed by an operator (the bastard!)
	 */
	if (do_hook(DISCONNECT_LIST, "Killed by %s (%s)", bastard, reason))
	{
		say("You have been killed by that fascist [%s] %s", 
			bastard, reason);
#ifdef QUIT_ON_OPERATOR_KILL
		say("Too bad, you lose.");
		irc_exit(1, NULL);
#endif
	}
	else if (background)
	{
		say("Too bad, you lose.");
		irc_exit(1, NULL);
	}

	/*
	 * Schedule a time-delayed auto-reconnect for non-bots only!
	 */
	if (!background && get_int_var(AUTO_RECONNECT_VAR))
	{
		char *	sc = m_sprintf("%d", from_server);
		add_timer(0, empty_string, 
			  (get_int_var(AUTO_RECONNECT_DELAY_VAR) * 1000), 1,
			  auto_reconnect_callback,
			  sc, NULL, get_current_winref());
	}

	window_check_servers();
}

static void 	p_ping (char *from, char **ArgList)
{
        PasteArgs(ArgList, 0);
	send_to_server("PONG %s", ArgList[0]);
}

static void 	p_silence (char *from, char **ArgList)
{
	char *target = ArgList[0];
	char mag = *target++;

	if (do_hook(SILENCE_LIST, "%c %s", mag, target))
	{
		if (mag == '+')
			say ("You will no longer recieve msgs from %s", target);
		else if (mag == '-')
			say ("You will now recieve msgs from %s", target);
		else
			say ("Unrecognized silence argument: %s", target);
	}
}


static void 	p_nick (char *from, char **ArgList)
{
	int		been_hooked = 0,
			its_me = 0;
	int		i, num;
	char		*line;
	char		*high;
	Channel_List	chanlist;
	int		ignored = 0;

	line = ArgList[0];

	/*
	 * Is this me changing nick?
	 */
	if (is_me(-1, from))
	{
		accept_server_nickname(from_server, line);
		its_me = 1;
		nick_command_is_pending(from_server, 0);
	}

	/* Um. Always do these things */
	notify_mark(from, 0, 0);
	channel_nick_rename(from_server, from, line);
	notify_mark(line, 1, 0);

	switch (check_ignore(from, FromUserHost, IGNORE_NICKS))
	{
		case IGNORED:
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
			break;
	}

	if (!check_flooding(from, NICK_FLOOD, line))
		return;

	if ((num = channel_getlist(from_server, line, chanlist)))
	{
	    for (i = 0; i < num; i++)
	    {
		if (check_ignore_channel(from, FromUserHost, chanlist[i],
						IGNORE_NICKS) == IGNORED)
		{
			ignored = 1;
			continue;
		}

		set_display_target(chanlist[i], LOG_CRAP);
		if (!do_hook(CHANNEL_NICK_LIST, "%s %s %s",
				chanlist[i], from, line))
			been_hooked = 1;
	    }
	}

	if (!been_hooked)
	{
		if (its_me)
			reset_display_target();
		else
			set_display_target(channel_what(from), LOG_CRAP);

		if (do_hook(NICKNAME_LIST, "%s %s", from, line))
			say("%s is now known as %s", from, line);
	}

	reset_display_target();
}

static void 	p_mode (char *from, char **ArgList)
{
	char	*channel;
	char	*line;

	if (!ArgList[0] || !ArgList[1])
		return;

	channel = ArgList[0];
	line = ArgList[1];

	PasteArgs(ArgList, 1);
	if (is_channel(ArgList[0]))
		channel_mode_update(ArgList[0], from_server, ArgList[1]);
	else
		update_user_mode(ArgList[1]);

	if (check_ignore_channel(from, FromUserHost, 
				 channel, IGNORE_CRAP) == IGNORED)
		return;

	set_display_target(channel, LOG_CRAP);
	if (get_int_var(MODE_STRIPPER_VAR))
		strip_modes(from, channel, line);

	if (is_channel(channel))
	{
		if (do_hook(MODE_LIST, "%s %s %s", 
				from, channel_friendly(channel, -1), line))
			say("Mode change \"%s\" on channel %s by %s",
				line, channel_friendly(channel, -1), from);
	}
	else
	{
		if (do_hook(MODE_LIST, "%s %s %s", from, channel, line))
			say("Mode change \"%s\" for user %s by %s",
				line, channel, from);
	}

	reset_display_target();
	update_all_status();
	return;
}

static void 	strip_modes (char *from, char *channel, char *line)
{
	char	*mode;
	char 	*pointer;
	char	mag = '+'; 		/* XXX Bogus */
        char    *copy = (char *) 0;
	char	*free_copy;

	copy = free_copy = LOCAL_COPY(line);
	mode = next_arg(copy, &copy);

	if (is_channel(channel))
	{
	    for (pointer = mode; *pointer; pointer++)
	    {
		char	c = *pointer;

		switch (c) 
		{
			case '+' :
			case '-' : 
				mag = c; 
				break;

			case 'l' : 
			{
				if (mag == '+')
					do_hook(MODE_STRIPPED_LIST,
						"%s %s %c%c %s",
						from, channel, mag,
						c,next_arg(copy,&copy));
				else
					do_hook(MODE_STRIPPED_LIST,
						"%s %s %c%c",
						from,channel,mag,c);
				break;
			}

			case 'a' : case 'i' : case 'm' : case 'n' :
			case 'p' : case 's' : case 't' : case 'z' : 
			case 'c' : case 'r' : case 'R' :
			{
				do_hook(MODE_STRIPPED_LIST,
					"%s %s %c%c",
					from, channel, mag, c);
				break;
			}

			case 'b' : case 'k' : case 'o' : case 'v' : 
			case 'e' : case 'I' :
			{
				do_hook(MODE_STRIPPED_LIST,
					"%s %s %c%c %s", 
					from, channel, mag, c,
					next_arg(copy,&copy));
				break;
			}
		}
	    }
	}

	else /* User mode */
	{
	    for (pointer = mode; *pointer; pointer++)
	    {
		char	c = *pointer;

		switch (c) 
		{
			case '+' :
			case '-' : mag = c; break;
#if 0	/* Not implemented yet */
			case 's' :
			{
				if (umode_s_takes_arg(from_server))
				{
					do_hook(MODE_STRIPPED_LIST, 
						"%s %s %c%c %s", 
						from, channel, mag, c, 
						next_arg(copy, &copy));
					break;
				}
				/* ELSE FALLTHROUGH */
			}
#endif
			default  : 
				do_hook(MODE_STRIPPED_LIST, 
					"%s %s %c%c", 
					from, channel, mag, c);
				break;
		}
	    }
	}
}

static void 	p_kick (char *from, char **ArgList)
{
	char	*channel,
		*who,
		*comment;

	if (!ArgList[0] || !ArgList[1])
		return;

	channel	= ArgList[0];
	who 	= ArgList[1];
	comment = ArgList[2] ? ArgList[2] : "(no comment)" ;

	/* I have been kicked from one of my channels */
	if (is_me(-1, who))
	{
		channel_status_change(channel, from_server, KICKED);
		update_all_status();
		set_display_target(channel, LOG_CRAP);
		if (do_hook(KICK_LIST, "%s %s %s %s", 
				who, from, 
				channel_friendly(channel, -1), comment))
			say("You have been kicked off channel %s by %s (%s)", 
				channel_friendly(channel, -1), from, comment);
		reset_display_target();
	}

	/* Some other poor slob has gotten the boot. */
	else
	{
		if (check_ignore_channel(from, FromUserHost, channel, 
					 IGNORE_CRAP) != IGNORED)
		{
		    set_display_target(channel, LOG_CRAP);
		    if (do_hook(KICK_LIST, "%s %s %s %s", 
				who, from, channel, comment))
			say("%s has been kicked off channel %s by %s (%s)", 
				who, channel_friendly(channel, -1), 
				from, comment);
		    reset_display_target();
		}

		channel_nick_remove(channel, from_server, who);
	}
}

static void 	p_part (char *from, char **ArgList)
{
	char	*channel = ArgList[0];

	if (is_me(-1, from))
		channel_remove(channel, from_server);
	else
		channel_nick_remove(channel, from_server, from);

	if ((check_ignore_channel(from, FromUserHost, 
			channel, IGNORE_PARTS) == IGNORED))
		return;

	set_display_target(channel, LOG_CRAP);	/* Channel don't exist dummy! */
	if (ArgList[1])		/* Dalnet part messages */
	{
		PasteArgs(ArgList, 1);
		if (do_hook(LEAVE_LIST, "%s %s %s %s", 
			from, channel, FromUserHost, ArgList[1]))
		    say("%s has left channel %s because (%s)", 
			from, channel_friendly(channel, -1), ArgList[1]);
	}
	else
	{
		if (do_hook(LEAVE_LIST, "%s %s %s", 
			from, channel, FromUserHost))
		    say("%s has left channel %s", 
			from, channel_friendly(channel, -1));
	}
	reset_display_target();
}

/*
 * Egads. i hope this is right.
 */
static void 	p_rpong (char *from, char **ArgList)
{
	if (!ArgList[3])
	{
		/* 
		 * We should always get an ArgList[3].  Punt if we dont.
		 */
		PasteArgs(ArgList, 0);
		say("RPONG %s (from %s)", ArgList[0], from);
	}
	else
	{
		/*
		 * :server RPONG yournick remoteserv ms :yourargs
		 *
		 * ArgList[0] -- our nickname (presumably)
		 * ArgList[1] -- The server we RPING'd
		 * ArgList[2] -- The number of ms it took to return
		 * ArgList[3] -- The arguments we passed (presumably)
		 */
		time_t delay = time(NULL) - atol(ArgList[3]);

		say("Pingtime %s - %s : %s ms (total delay: %ld s)",
			from, ArgList[1], ArgList[2], delay);
	}
}


static void 	rfc1459_odd (char *from, char *comm, char **ArgList)
{
	PasteArgs(ArgList, 0);
	if (do_hook(ODD_SERVER_STUFF_LIST, "%s %s %s", 
		from ? from : "*", comm, ArgList[0]))
	{
	    if (from)
		say("Odd server stuff: \"%s %s\" (%s)", comm, ArgList[0], from);
	    else
		say("Odd server stuff: \"%s %s\"", comm, ArgList[0]);
	}
}

protocol_command rfc1459[] = {
{	"ADMIN",	NULL,		NULL,		0		},
{	"AWAY",		NULL,		NULL,		0		},
{ 	"CONNECT",	NULL,		NULL,		0		},
{	"ERROR",	p_error,	NULL,		0		},
{	"ERROR:",	p_error,	NULL,		0		},
{	"INVITE",	p_invite,	NULL,		0		},
{	"INFO",		NULL,		NULL,		0		},
{	"ISON",		NULL,		NULL,		0		},
{	"JOIN",		p_channel,	NULL,		0		},
{	"KICK",		p_kick,		NULL,		0		},
{	"KILL",		p_kill,		NULL,		0		},
{	"LINKS",	NULL,		NULL,		0		},
{	"LIST",		NULL,		NULL,		0		},
{	"MODE",		p_mode,		NULL,		0		},
{	"NAMES",	NULL,		NULL,		0		},
{	"NICK",		p_nick,		NULL,		PROTO_NOQUOTE	},
{	"NOTICE",	parse_notice,	NULL,		0		},
{	"OPER",		NULL,		NULL,		0		},
{	"PART",		p_part,		NULL,		0		},
{	"PASS",		NULL,		NULL,		0 		},
{	"PING",		p_ping,		NULL,		0		},
{	"PONG",		p_pong,		NULL,		0		},
{	"PRIVMSG",	p_privmsg,	NULL,		0		},
{	"QUIT",		p_quit,		NULL,		PROTO_DEPREC	},
{	"REHASH",	NULL,		NULL,		0		},
{	"RESTART",	NULL,		NULL,		0		},
{	"RPONG",	p_rpong,	NULL,		0		},
{	"SERVER",	NULL,		NULL,		PROTO_NOQUOTE	},
{	"SILENCE",	p_silence,	NULL,		0		},
{	"SQUIT",	NULL,		NULL,		0		},
{	"STATS",	NULL,		NULL,		0		},
{	"SUMMON",	NULL,		NULL,		0		},
{	"TIME",		NULL,		NULL,		0		},
{	"TRACE",	NULL,		NULL,		0		},
{	"TOPIC",	p_topic,	NULL,		0		},
{	"USER",		NULL,		NULL,		0		},
{	"USERHOST",	NULL,		NULL,		0		},
{	"USERS",	NULL,		NULL,		0		},
{	"VERSION",	NULL,		NULL,		0		},
{	"WALLOPS",	p_wallops,	NULL,		0		},
{	"WHO",		NULL,		NULL,		0		},
{	"WHOIS",	NULL,		NULL,		0		},
{	"WHOWAS",	NULL,		NULL,		0		},
{	NULL,		NULL,		NULL,		0		}
};
#define NUMBER_OF_COMMANDS (sizeof(rfc1459) / sizeof(protocol_command)) - 2;
int 	num_protocol_cmds = -1;

/*
 * parse_server: parses messages from the server, doing what should be done
 * with them 
 */
void 	parse_server (char *orig_line)
{
	char *		from;
	char *		comm;
	int		numeric;
	char **		ArgList;
	char *		TrueArgs[MAXPARA + 1];
	protocol_command *retval;
	int		loc;
	int		cnt;
	const char *	old_fuh;
	char *		line;
	size_t		size;

	if (num_protocol_cmds == -1)
		num_protocol_cmds = NUMBER_OF_COMMANDS;

	if (!orig_line || !*orig_line)
		return;		/* empty line from server -> NOOP */

	if (*orig_line == ':')
	{
		if (!do_hook(RAW_IRC_LIST, "%s", orig_line + 1))
			return;
	}
	else if (!do_hook(RAW_IRC_LIST, "* %s", orig_line))
		return;

	if (inbound_line_mangler)
	{
		size = strlen(orig_line) * 3;
		line = alloca(size + 1);
		strcpy(line, orig_line);
		if (mangle_line(line, inbound_line_mangler, size) > size)
			yell("mangle_line truncated its result.  Ack.");
	}
	else
		line = orig_line;

	old_fuh = FromUserHost;
	ArgList = TrueArgs;
	BreakArgs(line, &from, ArgList);

	/*
	 * All RFC 1459 protocol messages are expected to have a sender
	 * (which may be empty), a command, and at least one argument
	 * (usually the target of the command).  Everything else is a
	 * broken command and should not be parsed.
	 */
	if ((comm = *ArgList++) && from && *ArgList)
	{
		if ((numeric = atoi(comm)))
			numbered_command(from, numeric, ArgList);
		else
		{
			retval = (protocol_command *)find_fixed_array_item(
				(void *)rfc1459, sizeof(protocol_command), 
				num_protocol_cmds + 1, comm, &cnt, &loc);

			if (cnt < 0 && rfc1459[loc].inbound_handler)
				rfc1459[loc].inbound_handler(from, ArgList);
			else
				rfc1459_odd(from, comm, ArgList);
		}
	}

	FromUserHost = old_fuh;
	from_server = -1;
}
