/* $EPIC:$ */
/*
 * symtab.c -- unified symbol table for epic
 *
 * Copyright  1999-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. All redistributions, whether of source code or in binary form must
 *    retain the the above copyright notice, the above paragraph (the one
 *    permitting redistribution), this list of conditions, and the following
 *    disclaimer.
 * 2. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */
#include "irc.h"
#include "alist.h"
#include "ircaux.h"
#include "output.h"
#include "symtab.h"

#if 0
typedef	struct 
{
	char *		name;
	void *		command;
	void *		alias;
} da_item;
#endif

typedef struct
{
	da_item **	list;
	int		max;
	int		max_alloc;
	alist_func	func;
	da_item **	cache;
	int		cache_size;
} da_set;

static da_set	command_list = { NULL, 0, 0, strncmp, NULL, 0 };
static void 	resize_cache (da_set *set, int newsize);

static	int	cache_hits = 0;
static	int	cache_misses = 0;
static	int	cache_passes = 0;

int	add_builtin_command_entry (const char *name, void *cmd)
{
	da_item *item;
	int	cnt;
	int	loc;

	/*
	 * Do brute force lookup
	 */
	item = (da_item *)find_array_item((array *)&command_list,
		name, &cnt, &loc);
	if (!item || cnt >= 0)
	{
		NEW(item)
		item->name = m_strdup(name);
		item->command = cmd;
		item->alias = NULL;
		add_to_array((array *)&command_list, (array_item *)&item->name);
	}
	else
		item->command = cmd;
	return 0;
}

int	add_alias_command_entry (const char *name, void *alias)
{
	da_item *item;
	int	cnt;
	int	loc;

	item = (da_item *)find_array_item((array *)&command_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
	{
		NEW(item)
		item->name = m_strdup(name);
		item->command = NULL;
		item->alias = alias;
		add_to_array((array *)&command_list, (array_item *)&item->name);
	}
	else
		item->alias = alias;
	return 0;
}

int	get_cmd_location (const char *name, int *count)
{
	da_item	*	item;
	int		location = -1;

	/* Ok.  Do a brute force lookup */
	if ((item = (da_item *)find_array_item((array *)&command_list,
					name, count, &location)))
	{
		if (*count > 1 || *count == 0)
		{
			cache_passes++;
			return -1;
		}
		return location;
	}
	return -1;
}

da_item *	get_cmd_item (const char *name, int *count, char **cname)
{
	da_item *item;
	int	location;
	int	i, j;

	if (cname)
		*cname = NULL;		/* Take _that_! */

	/*
	 * Do cache lookup first
	 */
        if (command_list.cache_size == 0)
                resize_cache(&command_list, 4);

	for (i = 0; i < command_list.cache_size; i++)
	{
		if (command_list.cache[i] && 
			command_list.cache[i]->name &&
			!strcmp(name, command_list.cache[i]->name))
		{
			item = command_list.cache[i];
			*count = -1;
			cache_hits++;
			goto cache_it;
		}

		/* XXXX this is bad cause this is free()d! */
		if (command_list.cache[i] && !command_list.cache[i]->name)
			command_list.cache[i] = NULL;
	}

	cache_misses++;
	if ((location = get_cmd_location(name, count)) == -1)
		return NULL;

	item = command_list.list[location];

cache_it:
	if (i != 0)
	{
		if (i == command_list.cache_size)
			i = command_list.cache_size - 1;

		/* Ok.  We found something.  Cache it. */
		for (j = i; j > 0; j--)
			command_list.cache[j] = command_list.cache[j-1];
		command_list.cache[0] = item;
	}

	return item;
}

void *	remove_builtin_command_entry (const char *name)
{
	da_item *item;
	int	cnt;
	int	loc;
	int	i;
	void *	retval;

	item = (da_item *)find_array_item((array *)&command_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
		return NULL;

	retval = (void *)item->command;
	item->command = NULL;
	if (!item->command && !item->alias)
	{
		/* Remove from cache */
		for (i = 0; i < command_list.cache_size; i++)
		{
			if (item == command_list.cache[i])
				command_list.cache[i] = NULL;
		}

		/* Remove from main list */
		item = (da_item *)remove_from_array((array *)&command_list, 
							name);
		new_free(&item->name);
		new_free((char **)&item);
	}
	return retval;
}

void *	remove_alias_command_entry (const char *name)
{
	da_item *item;
	int	cnt;
	int	loc;
	int	i;
	void *	retval;

	item = (da_item *)find_array_item((array *)&command_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
		return NULL;

	retval = (void *)item->alias;
	item->alias = NULL;
	if (!item->command && !item->alias)
	{
		/* Remove from cache */
		for (i = 0; i < command_list.cache_size; i++)
		{
			if (item == command_list.cache[i])
				command_list.cache[i] = NULL;
		}

		/* Remove from main list */
		item = (da_item *)remove_from_array((array *)&command_list, 
							name);
		new_free(&item->name);
		new_free((char **)&item);
	}
	return retval;
}

#define FOREACH_DA_TYPE(x)						\
void *	foreach_ ## x ## _item (int *count)				\
{									\
	while (*count < command_list.max && 				\
	       command_list.list[*count]-> x == NULL)			\
			(*count)++;					\
									\
	if (*count == command_list.max)					\
		return NULL;						\
									\
	return command_list.list[*count]-> x;				\
}

FOREACH_DA_TYPE(alias)


/**************************************************************************/
static da_set	func_list = { NULL, 0, 0, strncmp, NULL, 0 };

int	add_builtin_func_entry (const char *name, void *cmd)
{
	da_item *item;
	int	cnt;
	int	loc;

	/*
	 * Do brute force lookup
	 */
	item = (da_item *)find_array_item((array *)&func_list,
		name, &cnt, &loc);
	if (!item || cnt >= 0)
	{
		NEW(item)
		item->name = m_strdup(name);
		item->command = cmd;
		item->alias = NULL;
		add_to_array((array *)&func_list, (array_item *)&item->name);
	}
	else
		item->command = cmd;
	return 0;
}

int	add_alias_func_entry (const char *name, void *alias)
{
	da_item *item;
	int	cnt;
	int	loc;

	item = (da_item *)find_array_item((array *)&func_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
	{
		NEW(item)
		item->name = m_strdup(name);
		item->command = NULL;
		item->alias = alias;
		add_to_array((array *)&func_list, (array_item *)&item->name);
	}
	else
		item->alias = alias;
	return 0;
}

da_item *	get_func_item (const char *name, int *count, char **cname)
{
	da_item *item;
	int	location;
	int	i, j;

	if (cname)
		*cname = NULL;		/* Take _that_! */

	/*
	 * Do cache lookup first
	 */
        if (func_list.cache_size == 0)
                resize_cache(&func_list, 4);

	for (i = 0; i < func_list.cache_size; i++)
	{
		if (func_list.cache[i] && 
			func_list.cache[i]->name &&
			!strcmp(name, func_list.cache[i]->name))
		{
			item = func_list.cache[i];
			*count = -1;
			cache_hits++;
			goto cache_it;
		}

		/* XXXX this is bad cause this is free()d! */
		if (func_list.cache[i] && 
			!func_list.cache[i]->name)
		{
			func_list.cache[i] = NULL;
		}
	}

	/* Ok.  Do a brute force lookup */
	if ((item = (da_item *)find_array_item((array *)&func_list,
					name, count, &location)))
	{
		if (*count > 1 || *count == 0)
		{
			cache_passes++;
			return NULL;
		}
	}
	cache_misses++;

cache_it:
	if (i != 0)
	{
		if (i == func_list.cache_size)
			i = func_list.cache_size - 1;
		/* Ok.  We found something.  Cache it. */
		for (j = i; j > 0; j--)
			func_list.cache[j] = func_list.cache[j-1];
		func_list.cache[0] = item;
	}

	return item;
}

void *	remove_builtin_func_entry (const char *name)
{
	da_item *item;
	int	cnt;
	int	loc;
	int	i;
	void *	retval;

	item = (da_item *)find_array_item((array *)&func_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
		return NULL;

	retval = (void *)item->command;
	item->command = NULL;
	if (!item->command && !item->alias)
	{
		/* Remove from cache */
		for (i = 0; i < func_list.cache_size; i++)
		{
			if (item == func_list.cache[i])
				func_list.cache[i] = NULL;
		}

		/* Remove from main list */
		item = (da_item *)remove_from_array((array *)&func_list, 
							name);
		new_free(&item->name);
		new_free((char **)&item);
	}
	return retval;
}

void *	remove_alias_func_entry (const char *name)
{
	da_item *item;
	int	cnt;
	int	loc;
	int	i;
	void *	retval;

	item = (da_item *)find_array_item((array *)&func_list,
						name, &cnt, &loc);
	if (!item || cnt >= 0)
		return NULL;

	retval = (void *)item->alias;
	item->alias = NULL;
	if (!item->command && !item->alias)
	{
		/* Remove from cache */
		for (i = 0; i < func_list.cache_size; i++)
		{
			if (item == func_list.cache[i])
				func_list.cache[i] = NULL;
		}

		/* Remove from main list */
		item = (da_item *)remove_from_array((array *)&func_list, 
							name);
		new_free(&item->name);
		new_free((char **)&item);
	}
	return retval;
}

/**************************************************************************/
#if 0
typedef	struct 
{
	char *		name;
	void *		assign;
	void *		builtin;
	void *		set;
	void *		environ;
} vt_item;
#endif

typedef struct
{
	vt_item **	list;
	int		max;
	int		max_alloc;
	alist_func	func;
	vt_item **	cache;
	int		cache_size;
} vt_set;

static vt_set	variable_list = { NULL, 0, 0, strncmp, NULL, 0 };
static void 	resize_vt_cache (vt_set *set, int newsize);

#define ADD_VA_TYPE(x)							\
int	add_ ## x ## _var_entry (const char *name, void *cmd)		\
{									\
	vt_item *item;							\
	int	cnt;							\
	int	loc;							\
									\
	/*								\
	 * Do brute force lookup					\
	 */								\
	item = (vt_item *)find_array_item((array *)&variable_list,	\
		name, &cnt, &loc);					\
	if (!item || cnt >= 0)						\
	{								\
		NEW(item)						\
		item->name = m_strdup(name);				\
		item->assign = NULL;					\
		item->builtin = NULL;					\
		item->set = NULL;					\
		item->environ = NULL;					\
		item-> x = cmd;						\
		add_to_array((array *)&variable_list, 			\
			     (array_item *)&item->name);		\
	}								\
	else								\
		item-> x = cmd;						\
									\
	return 0;							\
}

ADD_VA_TYPE(assign)
ADD_VA_TYPE(builtin)
ADD_VA_TYPE(set)
ADD_VA_TYPE(environ)

int	get_var_location (const char *name, int *count)
{
	vt_item	*	item;
	int		location = -1;

	/* Ok.  Do a brute force lookup */
	if ((item = (vt_item *)find_array_item((array *)&variable_list,
					name, count, &location)))
	{
		if (*count >= 0)
		{
			cache_passes++;
			return -1;
		}
		return location;
	}
	return -1;
}

vt_item *	get_var_item (const char *name)
{
	vt_item *item;
	int	location;
	int	i, j;
	int	count;

	/*
	 * Do cache lookup first
	 */
        if (variable_list.cache_size == 0)
                resize_vt_cache(&variable_list, 4);

	for (i = 0; i < variable_list.cache_size; i++)
	{
		if (variable_list.cache[i] && 
			variable_list.cache[i]->name &&
			!strcmp(name, variable_list.cache[i]->name))
		{
			item = variable_list.cache[i];
			count = -1;
			cache_hits++;
			goto cache_it;
		}

		/* XXXX this is bad cause this is free()d! */
		if (variable_list.cache[i] && !variable_list.cache[i]->name)
			variable_list.cache[i] = NULL;
	}

	cache_misses++;

	if ((location = get_var_location(name, &count)) == -1)
		return NULL;
	item = variable_list.list[location];

cache_it:
	if (i != 0)
	{
		if (i == variable_list.cache_size)
			i = variable_list.cache_size - 1;
		/* Ok.  We found something.  Cache it. */
		for (j = i; j > 0; j--)
			variable_list.cache[j] = variable_list.cache[j-1];
		variable_list.cache[0] = item;
	}

	return item;
}

#define FOREACH_VA_TYPE(x)						\
void *	foreach_ ## x ## _item (int *count)				\
{									\
	while (*count < variable_list.max && 				\
	       variable_list.list[*count]-> x == NULL)			\
			(*count)++;					\
									\
	if (*count == variable_list.max)				\
		return NULL;						\
									\
	return variable_list.list[*count]-> x;				\
}

FOREACH_VA_TYPE(assign)

#define REMOVE_VA_TYPE(x) 						\
void *	remove_ ## x ## _var_entry (const char *name)			\
{									\
	vt_item *item;							\
	void *	retval;							\
	int	cnt;							\
	int	loc;							\
	int	i;							\
									\
	item = (vt_item *)find_array_item((array *)&variable_list,	\
						name, &cnt, &loc);	\
	if (!item || cnt >= 0)						\
		return NULL;						\
									\
	retval = item-> x ;						\
	item-> x = NULL;						\
	if (!item->assign && !item->builtin && !item->set && !item->environ) \
	{								\
		/* Remove from cache */					\
		for (i = 0; i < variable_list.cache_size; i++)		\
		{							\
			if (item == variable_list.cache[i])		\
				variable_list.cache[i] = NULL;		\
		}							\
									\
		/* Remove from main list */				\
		item = (vt_item *)remove_from_array((array *)&variable_list,  \
							name);		\
		new_free(&item->name);					\
		new_free((char **)&item);				\
	}								\
	return retval;							\
}

REMOVE_VA_TYPE(assign)
REMOVE_VA_TYPE(builtin)
REMOVE_VA_TYPE(set)
REMOVE_VA_TYPE(environ)


/**************************************************************************/
void	set_up_symbols (void)
{
	void register_builtin_commands(void);
	void register_builtin_functions(void);
	register_builtin_commands();
	register_builtin_functions();
}

static void 	resize_cache (da_set *set, int newsize)
{
	int 	c;
	int 	oldsize = set->cache_size;

	set->cache_size = newsize;
	RESIZE(set->cache, da_item *, set->cache_size);
	for (c = oldsize; c < set->cache_size; c++)
		set->cache[c] = NULL;
}

static void 	resize_vt_cache (vt_set *set, int newsize)
{
	int 	c;
	int 	oldsize = set->cache_size;

	set->cache_size = newsize;
	RESIZE(set->cache, vt_item *, set->cache_size);
	for (c = oldsize; c < set->cache_size; c++)
		set->cache[c] = NULL;
}

void	command_table_stats (void)
{
	int	i;

	yell("Cache hits, misses, passes, [%d/%d/%d]", 
		cache_hits, cache_misses, cache_passes);

	for (i = 0; i < command_list.cache_size; i++)
	{
		if (command_list.cache[i] && command_list.cache[i]->name)
			yell("(%d) -- [%s]", i, command_list.cache[i]->name);
		else
			yell("(%d) -- NULL", i);
	}

	for (i = 0; i < func_list.cache_size; i++)
	{
		if (func_list.cache[i] && func_list.cache[i]->name)
			yell("(%d) -- [%s]", i, func_list.cache[i]->name);
		else
			yell("(%d) -- NULL", i);
	}

	for (i = 0; i < variable_list.cache_size; i++)
	{
		if (variable_list.cache[i] && variable_list.cache[i]->name)
			yell("(%d) -- [%s]", i, variable_list.cache[i]->name);
		else
			yell("(%d) -- NULL", i);
	}
}


/* Obviously i havent gotten much further than this. */
